from collections.abc import Sequence
from typing import Literal

import matplotlib.axes
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.patches import Patch, Rectangle

from ..profile import calculate_approximate_ci, chi2_quantile_to_ratio
from ..result import Result

# kwargs passed to `matplotlib.axes.Axes.errorbar` for plotting confidence levels
cis_visualization_settings = {
    "capsize": 5,
    "linewidth": 2,
}


def profile_cis(
    result: Result,
    confidence_level: float = 0.95,
    df: int = 1,
    profile_indices: Sequence[int] = None,
    profile_list: int = 0,
    color: str | tuple = "C0",
    show_bounds: bool = False,
    ax: matplotlib.axes.Axes = None,
) -> matplotlib.axes.Axes:
    """
    Plot approximate confidence intervals based on profiles.

    Parameters
    ----------
    result:
        The result object after profiling.
    confidence_level:
        The confidence level in (0,1), which is translated to an approximate
        threshold assuming a chi2 distribution, using
        `pypesto.profile.chi2_quantile_to_ratio`.
    df:
        Degrees of freedom of the chi2 distribution.
    profile_indices:
        List of integer values specifying which profiles should be plotted.
        Defaults to the indices for which profiles were generated in profile
        list `profile_list`.
    profile_list:
        Index of the profile list to be used.
    color:
        Main plot color.
    show_bounds:
        Whether to show, and extend the plot to, the lower and upper bounds.
    ax:
        Axes object to use. Default: Create a new one.
    """
    # extract problem
    problem = result.problem
    # extract profile list
    profile_list = result.profile_result.list[profile_list]

    if profile_indices is None:
        profile_indices = [ix for ix, res in enumerate(profile_list) if res]

    if ax is None:
        _, ax = plt.subplots()

    confidence_ratio = chi2_quantile_to_ratio(confidence_level, df=df)

    # calculate intervals
    intervals = []
    for i_par in range(problem.dim_full):
        if i_par not in profile_indices:
            continue
        xs = profile_list[i_par].x_path[i_par]
        ratios = profile_list[i_par].ratio_path
        lb, ub = calculate_approximate_ci(
            xs=xs, ratios=ratios, confidence_ratio=confidence_ratio
        )
        intervals.append((lb, ub))

    x_names = [problem.x_names[ix] for ix in profile_indices]

    for ix, (lb, ub) in enumerate(intervals):
        half = (ub - lb) / 2
        ax.errorbar(
            lb + half,
            ix + 1,
            xerr=half,
            color=color,
            **cis_visualization_settings,
        )

    parameters_ind = np.arange(1, len(intervals) + 1)
    ax.set_yticks(parameters_ind)
    ax.set_yticklabels(x_names)
    ax.set_ylabel("Parameter")
    ax.set_xlabel("Parameter value")

    if show_bounds:
        lb = problem.lb_full[profile_indices]
        ax.plot(lb, parameters_ind, "k--", marker="+")
        ub = problem.ub_full[profile_indices]
        ax.plot(ub, parameters_ind, "k--", marker="+")

    return ax


def profile_nested_cis(
    result: Result,
    confidence_levels: Sequence[float] = (0.95, 0.9),
    df: int = 1,
    profile_indices: Sequence[int] = None,
    profile_list: int = 0,
    colors: Sequence = None,
    ax: matplotlib.axes.Axes = None,
    orientation: Literal["v", "h"] = "v",
):
    """
    Plot approximate nested confidence intervals based on profiles.

    Parameters
    ----------
    result:
        The result object with profiling results.
    confidence_levels:
        The confidence levels in (0,1), which are translated to an approximate
        threshold assuming a chi2 distribution, using
        `pypesto.profile.chi2_quantile_to_ratio`.
    df:
        Degrees of freedom of the chi2 distribution.
    profile_indices:
        List of integer values specifying which profiles should be plotted.
        Defaults to the indices for which profiles were generated in profile
        list `profile_list`.
    profile_list:
        Index of the profile list to be used.
    colors:
        A color for each confidence interval.
    ax:
        Axes object to use. Default: Create a new one.
    orientation:
        Orientation of the plot, either vertical or horizontal.
    """
    # extract problem
    problem = result.problem
    # extract profile list
    profile_list = result.profile_result.list[profile_list]

    n_cls = len(confidence_levels)
    ws = [(0.6 / n_cls) * i for i in range(1, n_cls + 1)]
    if colors is None:
        blues = cm.get_cmap("Blues")
        colors = [blues(i) for i in ws]

    # ensure that the confidence levels are sorted in decreasing order
    confidence_levels, colors = zip(
        *sorted(zip(confidence_levels, colors, strict=True), reverse=True),
        strict=True,
    )

    if profile_indices is None:
        profile_indices = [ix for ix, res in enumerate(profile_list) if res]

    if ax is None:
        _, ax = plt.subplots()

    legends = []
    for i, confidence_level in enumerate(confidence_levels):
        confidence_ratio = chi2_quantile_to_ratio(confidence_level, df=df)

        xs_list = []
        x = -ws[i] / 2
        rectangles = []
        for j, i_par in enumerate(profile_indices):
            conf_l_indices = [
                idx
                for idx, ratio in enumerate(profile_list[i_par].ratio_path)
                if ratio >= confidence_ratio
            ]
            xs = profile_list[i_par].x_path[i_par][conf_l_indices]
            xs_list.append(xs)

            par_ci = [np.min(xs), np.max(xs)]
            h = par_ci[1] - par_ci[0]

            if orientation == "v":
                rectangles.append(Rectangle((par_ci[0], x), h, ws[i]))
            else:
                rectangles.append(Rectangle((x, par_ci[0]), ws[i], h))
            x += 1

            # visualize parameter boundaries
            if orientation == "v":
                ax.plot(
                    [result.problem.lb_full[i_par]] * 2,
                    [j - 0.4, j + 0.4],
                    color="grey",
                )
                ax.plot(
                    [result.problem.ub_full[i_par]] * 2,
                    [j - 0.4, j + 0.4],
                    color="grey",
                )
            else:
                ax.plot(
                    [j - 0.4, j + 0.4],
                    [result.problem.lb_full[i_par]] * 2,
                    color="grey",
                )
                ax.plot(
                    [j - 0.4, j + 0.4],
                    [result.problem.ub_full[i_par]] * 2,
                    color="grey",
                )

        ax.add_collection(
            PatchCollection(
                rectangles, facecolors=colors[i], edgecolors="dimgrey"
            )
        )
        legends.append(
            Patch(color=colors[i], label=f"{confidence_level * 100}%")
        )

    x_names = [problem.x_names[ix] for ix in profile_indices]
    parameters_ind = np.arange(0, len(profile_indices))

    if orientation == "v":
        ax.set_yticks(parameters_ind)
        ax.set_yticklabels(x_names)
        ax.set_ylabel("Parameter")
        ax.set_xlabel("Parameter value")
    else:
        ax.set_xticks(parameters_ind)
        ax.set_xticklabels(ax.get_xticklabels(), ha="right")
        ax.set_xlabel("Parameter")
        ax.set_ylabel("Parameter value")

    ax.legend(
        title="Confidence level:",
        handles=legends,
        loc="upper center",
        bbox_to_anchor=(0.5, 1.16),
        ncol=len(legends),
    )

    return ax



"""Engines with multi-node parallelization."""

import logging
from typing import Any

import cloudpickle as pickle
from mpi4py import MPI
from mpi4py.futures import MPIPoolExecutor

from ..util import tqdm
from .base import Engine
from .task import Task

logger = logging.getLogger(__name__)


def work(pickled_task):
    """Unpickle and execute task."""
    task = pickle.loads(pickled_task)
    return task.execute()


class MPIPoolEngine(Engine):
    """
    Parallelize the task execution.

    Uses `mpi4py <https://mpi4py.readthedocs.io/en/stable/>`_.
    To be called with:
    ``mpiexec -np #Workers+1 python -m mpi4py.futures YOURFILE.py``
    """

    def __init__(self):
        super().__init__()

    def execute(
        self, tasks: list[Task], progress_bar: bool = None
    ) -> list[Any]:
        """
        Pickle tasks and distribute work to workers.

        Parameters
        ----------
        tasks:
            List of :class:`pypesto.engine.Task` to execute.
        progress_bar:
            Whether to display a progress bar.

        Returns
        -------
        A list of results.
        """
        pickled_tasks = [pickle.dumps(task) for task in tasks]

        n_procs = MPI.COMM_WORLD.Get_size()  # Size of communicator
        logger.info(
            f"Parallelizing on {n_procs - 1} workers with one manager."
        )

        with MPIPoolExecutor() as executor:
            results = executor.map(
                work, tqdm(pickled_tasks, enable=progress_bar)
            )
        return results



import copy
import logging
from collections.abc import Sequence

import numpy as np

from ..C import BETA_DECAY, EXPONENTIAL_DECAY
from ..problem import Problem
from ..result import McmcPtResult
from ..startpoint import PriorStartpoints
from ..util import tqdm
from .sampler import InternalSampler, Sampler

logger = logging.getLogger(__name__)


class ParallelTemperingSampler(Sampler):
    """Simple parallel tempering sampler.

    Parallel tempering is a Markov chain Monte Carlo (MCMC) method that
    uses multiple chains with different temperatures to sample from a
    probability distribution.
    The chains are coupled by swapping samples between them.
    This allows to sample from distributions with multiple modes more
    efficiently, as high-temperature chains can jump between modes, while
    low-temperature chains can sample the modes more precisely.

    This implementation is based on:

    * Vousden et al. 2016.
      Dynamic temperature selection for parallel tempering in Markov chain
      Monte Carlo simulations
      (https://doi.org/10.1093/mnras/stv2422),

    via a matlab-based reference implementation
    (https://github.com/ICB-DCM/PESTO/blob/master/private/performPT.m).
    """

    def __init__(
        self,
        internal_sampler: InternalSampler,
        betas: Sequence[float] = None,
        n_chains: int = None,
        options: dict = None,
    ):
        super().__init__(options)

        # set betas
        if (betas is None) == (n_chains is None):
            raise ValueError("Set either betas or n_chains.")
        if betas is None and self.options["beta_init"] == EXPONENTIAL_DECAY:
            logger.info('Initializing betas with "near-exponential decay".')
            betas = near_exponential_decay_betas(
                n_chains=n_chains,
                exponent=self.options["exponent"],
                max_temp=self.options["max_temp"],
            )
        elif betas is None and self.options["beta_init"] == BETA_DECAY:
            logger.info('Initializing betas with "beta decay".')
            betas = beta_decay_betas(
                n_chains=n_chains, alpha=self.options["alpha"]
            )
        if betas[0] != 1.0:
            raise ValueError("The first chain must have beta=1.0")
        self.betas0 = np.array(betas)
        self.betas = None

        self.temper_lpost = self.options["temper_log_posterior"]

        self.samplers = [
            copy.deepcopy(internal_sampler) for _ in range(len(self.betas0))
        ]
        # configure internal samplers
        for sampler in self.samplers:
            sampler.make_internal(temper_lpost=self.temper_lpost)

    @classmethod
    def default_options(cls) -> dict:
        """Return the default options for the sampler."""
        return {
            "max_temp": 5e4,
            "exponent": 4,
            "temper_log_posterior": False,
            "show_progress": None,
            "beta_init": BETA_DECAY,  # replaced in adaptive PT
            "alpha": 0.3,
            "warm_start_parallel_chains": 0.9,  # heuristic, not starting all chains at equal position
        }

    def initialize(self, problem: Problem, x0: np.ndarray | list[np.ndarray]):
        """Initialize all samplers."""
        n_chains = len(self.samplers)
        if isinstance(x0, list):
            x0s = x0
            if len(x0s) != n_chains and isinstance(x0s[0], np.ndarray):
                raise ValueError(
                    "If x0 is a list, its length must match the number of chains."
                )
        else:
            if self.options["warm_start_parallel_chains"] < 1.0:
                logger.info(
                    f"Initializing parallel chains with a combination of the starting point "
                    f"and prior samples with weight: {self.options['warm_start_parallel_chains']}."
                )
                get_start_params = PriorStartpoints(check_fval=True)
                x0_prior = get_start_params.sample(
                    n_starts=n_chains,
                    lb=problem.lb,
                    ub=problem.ub,
                    priors=problem.x_priors,
                )
                x0s = (
                    self.options["warm_start_parallel_chains"] * x0
                    + (1 - self.options["warm_start_parallel_chains"])
                    * x0_prior
                )
            else:
                x0s = [x0 for _ in range(n_chains)]
        for sampler, x0 in zip(self.samplers, x0s, strict=True):
            _problem = copy.deepcopy(problem)
            sampler.initialize(_problem, x0)
        self.betas = self.betas0

    def sample(self, n_samples: int, beta: float = 1.0):
        """Sample and swap in between samplers."""
        show_progress = self.options.get("show_progress", None)
        # loop over iterations
        for i_sample in tqdm(range(int(n_samples)), enable=show_progress):
            # TODO test
            # sample
            for sampler, beta in zip(self.samplers, self.betas, strict=True):
                sampler.sample(n_samples=1, beta=beta)

            # swap samples
            swapped = self.swap_samples()

            # adjust temperatures
            self.adjust_betas(i_sample, swapped)

    def get_samples(self) -> McmcPtResult:
        """Concatenate all chains."""
        results = [sampler.get_samples() for sampler in self.samplers]
        trace_x = np.array([result.trace_x[0] for result in results])
        trace_neglogpost = np.array(
            [result.trace_neglogpost[0] for result in results]
        )
        trace_neglogprior = np.array(
            [result.trace_neglogprior[0] for result in results]
        )
        return McmcPtResult(
            trace_x=trace_x,
            trace_neglogpost=trace_neglogpost,
            trace_neglogprior=trace_neglogprior,
            betas=self.betas,
        )

    def swap_samples(self) -> Sequence[bool]:
        """Swap samples as in Vousden2016."""
        # for recording swaps
        swapped = []

        if len(self.betas) == 1:
            # nothing to be done
            return swapped

        # beta differences
        dbetas = self.betas[:-1] - self.betas[1:]

        # loop over chains from highest temperature down
        for dbeta, sampler1, sampler2 in reversed(
            list(
                zip(dbetas, self.samplers[:-1], self.samplers[1:], strict=True)
            )
        ):
            # extract samples
            sample1 = sampler1.get_last_sample()
            sample2 = sampler2.get_last_sample()

            # extract log likelihood values
            sample1_llh = sample1.lpost - sample1.lprior
            sample2_llh = sample2.lpost - sample2.lprior

            # swapping probability
            p_acc_swap = dbeta * (sample2_llh - sample1_llh)

            # flip a coin
            u = np.random.uniform(0, 1)

            # check acceptance
            swap = np.log(u) < p_acc_swap
            if swap:
                # swap
                sampler2.set_last_sample(sample1)
                sampler1.set_last_sample(sample2)

            # record
            swapped.insert(0, swap)
        return swapped

    def adjust_betas(self, i_sample: int, swapped: Sequence[bool]):
        """Adjust temperature values. Default: Do nothing."""


def beta_decay_betas(n_chains: int, alpha: float) -> np.ndarray:
    """Initialize betas to the (j-1)th quantile of a Beta(alpha, 1) distribution.

    Proposed by Xie et al. (2011) to be used for thermodynamic integration.

    Parameters
    ----------
    n_chains:
        Number of chains to use.
    alpha:
        Tuning parameter that modulates the skew of the distribution over the temperatures.
        For alpha=1 we have a uniform distribution, and as alpha decreases towards zero,
        temperatures become positively skewed. Xie et al. (2011) propose alpha=0.3 as a good start.
    """
    if alpha <= 0 or alpha > 1:
        raise ValueError("alpha must be in (0, 1]")

    # special case of one chain
    if n_chains == 1:
        return np.array([1.0])

    return np.power(np.arange(n_chains) / (n_chains - 1), 1 / alpha)[::-1]


def near_exponential_decay_betas(
    n_chains: int, exponent: float, max_temp: float
) -> np.ndarray:
    """Initialize betas in a near-exponential decay scheme.

    Parameters
    ----------
    n_chains:
        Number of chains to use.
    exponent:
        Decay exponent. The higher, the more small temperatures are used.
    max_temp:
        Maximum chain temperature.
    """
    # special case of one chain
    if n_chains == 1:
        return np.array([1.0])

    temperatures = (
        np.linspace(1, max_temp ** (1 / exponent), n_chains) ** exponent
    )
    betas = 1 / temperatures

    return betas



import inspect
import warnings
from collections.abc import Sequence
from copy import deepcopy
from typing import Any

import numpy as np

from ..C import (
    FVAL,
    GRAD,
    HESS,
    HESSP,
    RDATAS,
    RES,
    SRES,
    ModeType,
)
from .base import ObjectiveBase, ResultDict


class AggregatedObjective(ObjectiveBase):
    """Aggregates multiple objectives into one objective."""

    def __init__(
        self,
        objectives: Sequence[ObjectiveBase],
        x_names: Sequence[str] = None,
    ):
        """
        Initialize objective.

        Parameters
        ----------
        objectives:
            Sequence of pypesto.ObjectiveBase instances
        x_names:
            Sequence of names of the (optimized) parameters.
            (Details see documentation of x_names in
            :class:`pypesto.ObjectiveBase`)
        """
        # input typechecks
        if not isinstance(objectives, Sequence):
            raise TypeError(
                f"Objectives must be a Sequence, was {type(objectives)}."
            )

        if not all(
            isinstance(objective, ObjectiveBase) for objective in objectives
        ):
            raise TypeError(
                "Objectives must only contain elements of type"
                "pypesto.Objective"
            )

        if not objectives:
            raise ValueError("Length of objectives must be at least one")

        self._objectives = objectives

        super().__init__(x_names=x_names)

    def __deepcopy__(self, memodict=None):
        """Create copy of objective."""
        other = AggregatedObjective(
            objectives=[deepcopy(objective) for objective in self._objectives],
            x_names=deepcopy(self.x_names),
        )
        for key in set(self.__dict__.keys()) - {"_objectives", "x_names"}:
            other.__dict__[key] = deepcopy(self.__dict__[key])

        return other

    def check_mode(self, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        return all(
            objective.check_mode(mode) for objective in self._objectives
        )

    def check_sensi_orders(
        self,
        sensi_orders: tuple[int, ...],
        mode: ModeType,
    ) -> bool:
        """See `ObjectiveBase` documentation."""
        return all(
            objective.check_sensi_orders(sensi_orders, mode)
            for objective in self._objectives
        )

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: tuple[int, ...],
        mode: ModeType,
        kwargs_list: Sequence[dict[str, Any]] = None,
        return_dict: bool = False,
        **kwargs,
    ) -> ResultDict:
        """
        See `ObjectiveBase` for more documentation.

        Main method to overwrite from the base class. It handles and
        delegates the actual objective evaluation.

        Parameters
        ----------
        kwargs_list:
            Objective-specific keyword arguments, where the dictionaries are
            ordered by the objectives.
        """
        if kwargs_list is None:
            kwargs_list = [{}] * len(self._objectives)
        elif len(kwargs_list) != len(self._objectives):
            raise ValueError(
                "The length of `kwargs_list` must match the number of "
                "objectives you are aggregating."
            )
        for objective_, objective_kwargs in zip(
            self._objectives, kwargs_list, strict=True
        ):
            if (
                "return_dict"
                in inspect.signature(objective_.call_unprocessed).parameters
            ):
                objective_kwargs["return_dict"] = return_dict
            else:
                warnings.warn(
                    "Please add `return_dict` to the argument list of your "
                    "objective's `call_unprocessed` method. "
                    f"Current objective: `{type(objective_)}`.",
                    DeprecationWarning,
                    stacklevel=1,
                )
        return aggregate_results(
            [
                objective.call_unprocessed(
                    x,
                    sensi_orders,
                    mode,
                    **kwargs,
                    **cur_kwargs,
                )
                for objective, cur_kwargs in zip(
                    self._objectives, kwargs_list, strict=True
                )
            ]
        )

    def initialize(self):
        """See `ObjectiveBase` documentation."""
        for objective in self._objectives:
            objective.initialize()

    def get_config(self) -> dict:
        """Return basic information of the objective configuration."""
        info = super().get_config()
        for n_obj, obj in enumerate(self._objectives):
            info[f"objective_{n_obj}"] = obj.get_config()
        return info


def aggregate_results(rvals: Sequence[ResultDict]) -> ResultDict:
    """
    Aggregate the results from the provided ResultDicts into a single one.

    Parameters
    ----------
    rvals:
        results to aggregate
    """
    # sum over fval/grad/hess, if available in all rvals
    result = {
        key: sum(rval[key] for rval in rvals)
        for key in [FVAL, GRAD, HESS, HESSP]
        if all(key in rval for rval in rvals)
    }

    # extract rdatas and flatten
    result[RDATAS] = []
    for rval in rvals:
        if RDATAS in rval:
            result[RDATAS].extend(rval[RDATAS])

    # initialize res and sres
    if RES in rvals[0]:
        res = np.asarray(rvals[0][RES])
    else:
        res = None

    if SRES in rvals[0]:
        sres = np.asarray(rvals[0][SRES])
    else:
        sres = None

    # skip iobj=0 after initialization, stack matrices
    for rval in rvals[1:]:
        if res is not None:
            res = np.hstack([res, np.asarray(rval[RES])])
        if sres is not None:
            sres = np.vstack([sres, np.asarray(rval[SRES])])

    # fill res, sres into result
    if res is not None:
        result[RES] = res
    if sres is not None:
        result[SRES] = sres

    return result



"""
Problem
=======

A problem contains the objective as well as all information like prior
describing the problem to be solved.

"""

from .base import Problem
from .hierarchical import HierarchicalProblem



"""
Hierarchical
============

Contains an implementation of the hierarchical optimization approach, which
decomposes the parameter estimation problem into an outer and an inner problem.
In the outer problem, only dynamic parameters are optimized.
In the inner problem, conditional on the outer solution, static parameters are
optimized.
Static parameters can be parameters affecting only the model observables,
such as scaling factors, offsets, and noise parameters. Further, they can be
spline parameters, which are used to approximate non-linear measurement mappings
of semi-quantitative data. Finally, they can be optimal scaling parameters,
which are used for the integration of ordinal data.

Hierarchical optimization has the advantage that the outer problem is typically
less complex than the full problem, and thus can be solved more efficiently.
Further, in the relative data case, the inner problem can be solved analytically,
which is more efficient.
Thus, hierarchical optimization can be used to speed up parameter estimation,
finding optimal values more efficiently and reliably.
"""

from .base_parameter import InnerParameter
from .base_problem import InnerProblem
from .base_solver import InnerSolver
from .inner_calculator_collector import InnerCalculatorCollector



"""Process a model selection :class:`ModelProblem` after calibration."""

import warnings
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from petab_select.constants import TYPE_PATH, Criterion

from .. import store, visualize
from .model_problem import TYPE_POSTPROCESSOR, ModelProblem

__all__ = [
    "model_id_binary_postprocessor",
    "multi_postprocessor",
    "report_postprocessor",
    "save_postprocessor",
    "waterfall_plot_postprocessor",
]


def multi_postprocessor(
    problem: ModelProblem,
    postprocessors: list[TYPE_POSTPROCESSOR] = None,
):
    """Combine multiple postprocessors into a single postprocessor.

    See :meth:`save_postprocessor` for usage hints.

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    postprocessors:
        A list of postprocessors, which will be sequentially applied to the
        optimized model ``problem``.
        The location where results will be stored.
    """
    for postprocessor in postprocessors:
        postprocessor(problem)


def waterfall_plot_postprocessor(
    problem: ModelProblem,
    output_path: TYPE_PATH = ".",
):
    """Produce a waterfall plot.

    See :meth:`save_postprocessor` for usage hints and argument documentation.
    """
    visualize.waterfall(problem.minimize_result)
    plot_output_path = Path(output_path) / (str(problem.model.hash) + ".png")
    plt.savefig(str(plot_output_path))


def save_postprocessor(
    problem: ModelProblem,
    output_path: TYPE_PATH = ".",
    use_model_hash: bool = False,
):
    """Save the parameter estimation result.

    When used, first set the output folder for results, e.g. with
    :func:`functools.partial`. This is because postprocessors should take only a
    single parameter: an optimized model.

    .. code-block:: python

       from functools import partial
       output_path = 'results'
       pp = partial(save_postprocessor, output_path=output_path)
       selector = pypesto.select.ModelSelector(
           problem=problem,
           model_postprocessor=pp,
       )

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    output_path:
        The location where output will be stored.
    use_model_hash:
        Whether the filename should use the model hash. Defaults to ``False``,
        in which case the model ID is used instead.
    """
    stem = problem.model.model_id
    if use_model_hash:
        stem = str(problem.model.hash)
    store.write_result(
        problem.minimize_result,
        Path(output_path) / (stem + ".hdf5"),
    )


def model_id_binary_postprocessor(problem: ModelProblem):
    """Change a PEtab Select model ID to a binary string.

    Changes the model ID in-place to be a string like ``M_ijk``, where
    ``i``, ``j``, ``k``, etc. are ``1`` if the parameter in that position is estimated,
    or ``0`` if the parameter is fixed.

    To ensure that other postprocessors (e.g. :func:`report_postprocessor`) use this
    new model ID, when in use with a :func:`multi_postprocessor`, ensure this is
    before the other postprocessors in the ``postprocessors`` argument of
    :func:`multi_postprocessor`.

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    """
    warnings.warn(
        (
            "This is obsolete. Model IDs are by default the model hash, which "
            "is now similar to the binary string."
        ),
        DeprecationWarning,
        stacklevel=2,
    )
    problem.model.model_id = str(problem.model.hash)


def report_postprocessor(
    problem: ModelProblem,
    output_filepath: TYPE_PATH,
    criteria: list[Criterion] = None,
):
    """Create a TSV table of model selection results.

    Parameters
    ----------
    problem:
        A model selection :class:`ModelProblem` that has been optimized.
    output_filepath:
        The file path where the report will be saved.
    criteria:
        The criteria that will be in the report. Defaults to nllh, AIC, AICc,
        and BIC.
    """
    output_filepath = Path(output_filepath)
    write_header = False
    # Only write the header if the file doesn't yet exist or is empty.
    if not output_filepath.exists() or output_filepath.stat().st_size == 0:
        write_header = True
    if criteria is None:
        criteria = [
            Criterion.NLLH,
            Criterion.AIC,
            Criterion.AICC,
            Criterion.BIC,
        ]

    start_optimization_times = problem.minimize_result.optimize_result.time

    header = []
    row = []

    header.append("model_id")
    row.append(problem.model.model_id)

    header.append("total_time")
    row.append(str(sum(start_optimization_times)))

    for criterion in criteria:
        header.append(criterion.value)
        row.append(str(problem.model.get_criterion(criterion)))

    # Arbitrary convergence criterion
    header.append("n_converged")
    row.append(
        str(
            (
                np.array(problem.minimize_result.optimize_result.fval)
                < (problem.minimize_result.optimize_result.list[0].fval + 0.1)
            ).sum()
        )
    )

    for start_index, start_optimization_time in enumerate(
        start_optimization_times
    ):
        header.append(f"start_time_{start_index}")
        row.append(str(start_optimization_time))

    with open(output_filepath, "a+") as f:
        if write_header:
            f.write("\t".join(header) + "\n")
        f.write("\t".join(row) + "\n")



from __future__ import annotations

import warnings

import numpy as np
from scipy.optimize import minimize

from ...C import (
    CURRENT_SIMULATION,
    DATAPOINTS,
    EXPDATA_MASK,
    INNER_NOISE_PARS,
    MAX_DATAPOINT,
    MIN_DATAPOINT,
    MIN_DIFF_FACTOR,
    MIN_SIM_RANGE,
    N_SPLINE_PARS,
    NUM_DATAPOINTS,
    OPTIMIZE_NOISE,
    REGULARIZATION_FACTOR,
    REGULARIZE_SPLINE,
    SCIPY_FUN,
    SCIPY_SUCCESS,
    SCIPY_X,
    InnerParameterType,
)
from ..base_solver import InnerSolver
from .parameter import SplineInnerParameter
from .problem import SemiquantProblem

try:
    from amici.sim._parameter_mapping import ParameterMapping
except ImportError:
    pass


class SemiquantInnerSolver(InnerSolver):
    """Solver of the inner subproblem of spline approximation for nonlinear-monotone data.

    Options
    -------
    min_diff_factor:
        Determines the minimum difference between two consecutive spline
        as ``min_diff_factor * (measurement_range) / n_spline_pars``.
        Default is 1/2.
    """

    def __init__(self, options: dict = None):
        self.options = {
            **self.get_default_options(),
            **(options or {}),
        }
        self.validate_options()

    def validate_options(self):
        """Validate the current options dictionary."""
        if not isinstance(self.options[MIN_DIFF_FACTOR], float):
            raise TypeError(f"{MIN_DIFF_FACTOR} must be of type float.")
        elif self.options[MIN_DIFF_FACTOR] < 0:
            raise ValueError(f"{MIN_DIFF_FACTOR} must not be negative.")

        elif not isinstance(self.options[REGULARIZE_SPLINE], bool):
            raise TypeError(f"{REGULARIZE_SPLINE} must be of type bool.")
        if self.options[REGULARIZE_SPLINE]:
            if not isinstance(self.options[REGULARIZATION_FACTOR], float):
                raise TypeError(
                    f"{REGULARIZATION_FACTOR} must be of type float."
                )
            elif self.options[REGULARIZATION_FACTOR] < 0:
                raise ValueError(
                    f"{REGULARIZATION_FACTOR} must not be negative."
                )

        for key in self.options:
            if key not in self.get_default_options():
                raise ValueError(f"Unknown SplineInnerSolver option {key}.")

    def solve(
        self,
        problem: SemiquantProblem,
        sim: list[np.ndarray],
        amici_sigma: list[np.ndarray],
    ) -> list:
        """Get results for every group (inner optimization problem).

        Parameters
        ----------
        problem:
            InnerProblem from pyPESTO hierarchical.
        sim:
            Simulations from AMICI.
        amici_sigma:
            List of sigmas from AMICI.

        Returns
        -------
        List of optimization results of the inner subproblem.
        """
        inner_results = []
        for group in problem.get_groups_for_xs(InnerParameterType.SPLINE):
            group_dict = problem.groups[group]
            group_dict[CURRENT_SIMULATION] = extract_expdata_using_mask(
                expdata=sim, mask=group_dict[EXPDATA_MASK]
            )

            # If the noise parameters are optimized in the outer problem,
            # extract them from amici return data.
            if not group_dict[OPTIMIZE_NOISE]:
                group_dict[INNER_NOISE_PARS] = extract_expdata_using_mask(
                    expdata=amici_sigma, mask=group_dict[EXPDATA_MASK]
                )[0]

            # Optimize the spline for this group.
            inner_result_for_group = self._optimize_spline(
                inner_parameters=problem.get_free_xs_for_group(group),
                group_dict=group_dict,
            )

            inner_results.append(inner_result_for_group)
            save_inner_parameters_to_inner_problem(
                inner_problem=problem,
                s=inner_result_for_group[SCIPY_X],
                group=group,
            )
        return inner_results

    @staticmethod
    def calculate_obj_function(x_inner_opt: list):
        """Calculate the inner objective function value.

        Calculates the inner objective function value from a list of inner
        optimization results returned from `_optimize_spline`.

        Parameters
        ----------
        x_inner_opt:
            List of optimization results

        Returns
        -------
        Inner objective function value.
        """
        if False in (
            x_inner_opt[idx][SCIPY_SUCCESS] for idx in range(len(x_inner_opt))
        ):
            obj = np.inf
            warnings.warn(
                "Inner optimization failed.",
                stacklevel=2,
            )
        else:
            obj = np.sum(
                [
                    x_inner_opt[idx][SCIPY_FUN]
                    for idx in range(len(x_inner_opt))
                ]
            )
        return obj

    def calculate_gradients(
        self,
        problem: SemiquantProblem,
        x_inner_opt: list[dict],
        sim: list[np.ndarray],
        amici_sigma: list[np.ndarray],
        sy: list[np.ndarray],
        amici_ssigma: list[np.ndarray],
        parameter_mapping: ParameterMapping,
        par_opt_ids: list,
        par_sim_ids: list,
        par_edatas_indices: list,
        snllh: np.ndarray,
    ):
        """Calculate gradients of the inner objective function.

        Calculates gradients of the objective function with respect to outer
        (dynamical) parameters.

        Parameters
        ----------
        problem:
            Optimal scaling inner problem.
        x_inner_opt:
            List of optimization results of the inner subproblem.
        sim:
            Model simulations.
        sigma:
            Model noise parameters.
        sy:
            Model sensitivities.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        par_opt_ids:
            Ids of outer otimization parameters.
        par_sim_ids:
            Ids of outer simulation parameters, includes fixed parameters.
        snllh:
            A zero-initialized vector of the same length as ``par_opt_ids`` to store the
            gradients in. Will be modified in-place.

        Returns
        -------
        The gradients with respect to the outer parameters.
        """
        already_calculated = set()

        for condition_map_sim_var in [
            cond_par_map.map_sim_var for cond_par_map in parameter_mapping
        ]:
            for par_sim, par_opt in condition_map_sim_var.items():
                if (
                    not isinstance(par_opt, str)
                    or par_opt in already_calculated
                ):
                    continue
                elif par_opt not in par_opt_ids:
                    continue
                else:
                    already_calculated.add(par_opt)
                par_sim_idx = par_sim_ids.index(par_sim)
                par_opt_idx = par_opt_ids.index(par_opt)
                grad = 0.0

                sy_for_outer_parameter = [
                    (
                        sy_cond[:, par_edata_indices.index(par_sim_idx), :]
                        if par_sim_idx in par_edata_indices
                        else np.zeros(sy_cond[:, 0, :].shape)
                    )
                    for sy_cond, par_edata_indices in zip(
                        sy, par_edatas_indices, strict=True
                    )
                ]
                ssigma_for_outer_parameter = [
                    (
                        ssigma_cond[:, par_edata_indices.index(par_sim_idx), :]
                        if par_sim_idx in par_edata_indices
                        else np.zeros(ssigma_cond[:, 0, :].shape)
                    )
                    for ssigma_cond, par_edata_indices in zip(
                        amici_ssigma, par_edatas_indices, strict=True
                    )
                ]

                for group_idx, group in enumerate(
                    problem.get_groups_for_xs(InnerParameterType.SPLINE)
                ):
                    # Get the reformulated spline parameters
                    s = np.asarray(x_inner_opt[group_idx][SCIPY_X])
                    group_dict = problem.groups[group]

                    measurements = group_dict[DATAPOINTS]
                    sigma = group_dict[INNER_NOISE_PARS]
                    sim_all = group_dict[CURRENT_SIMULATION]
                    N = group_dict[N_SPLINE_PARS]
                    K = group_dict[NUM_DATAPOINTS]

                    sy_all = extract_expdata_using_mask(
                        expdata=sy_for_outer_parameter,
                        mask=group_dict[EXPDATA_MASK],
                    )
                    ssigma_all = extract_expdata_using_mask(
                        expdata=ssigma_for_outer_parameter,
                        mask=group_dict[EXPDATA_MASK],
                    )

                    delta_c, c, n = self._rescale_spline_bases(
                        sim_all=sim_all, N=N, K=K
                    )
                    delta_c_dot, c_dot = calculate_spline_bases_gradient(
                        sim_all=sim_all, sy_all=sy_all, N=N
                    )

                    # For the reformulated problem, mu can be calculated
                    # as the inner gradient at the optimal point s.
                    mu = _calculate_nllh_gradient_for_group(
                        s=s,
                        sim_all=sim_all,
                        measurements=measurements,
                        N=N,
                        delta_c=delta_c,
                        c=c,
                        n=n,
                        regularization_factor=self.options[
                            REGULARIZATION_FACTOR
                        ],
                        regularize_spline=self.options[REGULARIZE_SPLINE],
                        group_dict=group_dict,
                    )
                    min_meas = group_dict[MIN_DATAPOINT]
                    max_meas = group_dict[MAX_DATAPOINT]
                    min_diff = self._get_minimal_difference(
                        measurement_range=max_meas - min_meas,
                        N=N,
                        min_diff_factor=self.options[MIN_DIFF_FACTOR],
                    )

                    # If the spline parameter is at its boundary, the
                    # corresponding Lagrangian multiplier mu is set to 0.
                    min_diff_all = np.full(N, min_diff)
                    min_diff_all[0] = 0.0
                    mu = np.asarray(
                        [
                            (
                                mu[i]
                                if np.isclose(s[i] - min_diff_all[i], 0)
                                else 0
                            )
                            for i in range(len(s))
                        ]
                    )

                    # Calculate the (dJ_dy * dy_dtheta) term:
                    dy_grad_term = calculate_dy_term(
                        sim_all=sim_all,
                        sy_all=sy_all,
                        measurements=measurements,
                        s=s,
                        N=N,
                        delta_c=delta_c,
                        delta_c_dot=delta_c_dot,
                        c=c,
                        c_dot=c_dot,
                        n=n,
                    )

                    # Calculate the (dJ_dsigma^2 * dsigma^2_dtheta) term:
                    if not group_dict[OPTIMIZE_NOISE]:
                        residual_squared = _calculate_residuals_for_group(
                            s=s,
                            sim_all=sim_all,
                            measurements=measurements,
                            N=N,
                            delta_c=delta_c,
                            c=c,
                            n=n,
                        )
                        dJ_dsigma2 = (
                            K / (2 * sigma**2) - residual_squared / sigma**4
                        )
                        dsigma2_dtheta = ssigma_all[0] * sigma
                        dsigma_grad_term = dJ_dsigma2 * dsigma2_dtheta
                    # If we optimize the noise hierarchically,
                    # the last term (dJ_dsigma^2 * dsigma^2_dtheta) is always 0
                    # since the sigma is optimized such that dJ_dsigma2=0.
                    else:
                        dsigma_grad_term = 0.0

                    # Combine all terms to get the complete gradient contribution
                    grad += dy_grad_term / sigma**2 + dsigma_grad_term

                snllh[par_opt_idx] = grad

        return snllh

    @staticmethod
    def get_default_options() -> dict:
        """Return default options for solving the inner problem."""
        options = {
            MIN_DIFF_FACTOR: 0.0,
            REGULARIZE_SPLINE: False,
            REGULARIZATION_FACTOR: 0.0,
        }
        return options

    def _optimize_spline(
        self,
        inner_parameters: list[SplineInnerParameter],
        group_dict: dict,
    ):
        """Run optimization for the inner problem.

        Parameters
        ----------
        inner_parameters:
            The spline inner parameters.
        group_dict:
            The group dictionary.
        """
        (
            distance_between_bases,
            spline_bases,
            intervals_per_sim,
        ) = self._rescale_spline_bases(
            sim_all=group_dict[CURRENT_SIMULATION],
            N=group_dict[N_SPLINE_PARS],
            K=group_dict[NUM_DATAPOINTS],
        )

        min_diff = self._get_minimal_difference(
            measurement_range=group_dict[MAX_DATAPOINT]
            - group_dict[MIN_DATAPOINT],
            N=group_dict[N_SPLINE_PARS],
            min_diff_factor=self.options[MIN_DIFF_FACTOR],
        )

        inner_options = self._get_inner_optimization_options(
            inner_parameters=inner_parameters,
            N=group_dict[N_SPLINE_PARS],
            min_meas=group_dict[MIN_DATAPOINT],
            max_meas=group_dict[MAX_DATAPOINT],
            min_diff=min_diff,
        )

        # Wrap the analytical optimization of sigma and
        # the regularization into the objective function
        def objective_function_wrapper(x):
            return _calculate_nllh_for_group(
                s=x,
                sim_all=group_dict[CURRENT_SIMULATION],
                measurements=group_dict[DATAPOINTS],
                N=group_dict[N_SPLINE_PARS],
                delta_c=distance_between_bases,
                c=spline_bases,
                n=intervals_per_sim,
                regularization_factor=self.options[REGULARIZATION_FACTOR],
                regularize_spline=self.options[REGULARIZE_SPLINE],
                group_dict=group_dict,
            )

        # Wrap the analytical optimization of sigma and
        # the regularization into the gradient function
        def inner_gradient_wrapper(x):
            return _calculate_nllh_gradient_for_group(
                s=x,
                sim_all=group_dict[CURRENT_SIMULATION],
                measurements=group_dict[DATAPOINTS],
                N=group_dict[N_SPLINE_PARS],
                delta_c=distance_between_bases,
                c=spline_bases,
                n=intervals_per_sim,
                regularization_factor=self.options[REGULARIZATION_FACTOR],
                regularize_spline=self.options[REGULARIZE_SPLINE],
                group_dict=group_dict,
            )

        results = minimize(
            objective_function_wrapper,
            jac=inner_gradient_wrapper,
            **inner_options,
        )

        return results

    @staticmethod
    def _rescale_spline_bases(sim_all: np.ndarray, N: int, K: int):
        """Rescale the spline bases.

        Before the optimization of the spline parameters, we have to fix the
        spline bases to some values. We choose to scale them to the current
        simulation. In case of simulations that are very close to each other,
        we choose to scale closely around the average value of the simulations,
        to avoid numerical problems (as we often divide by delta_c).

        Parameters
        ----------
        sim_all:
            The current simulation.
        N:
            The number of spline parameters.
        K:
            The number of simulations.

        Returns
        -------
        distance_between_bases:
            The distance between the spline bases.
        spline_bases:
            The rescaled spline bases.
        intervals_per_sim:
            List of indices of intervals each simulation belongs to.
        """
        min_idx = np.argmin(sim_all)
        max_idx = np.argmax(sim_all)

        min_all = sim_all[min_idx]
        max_all = sim_all[max_idx]

        n = np.ones(K)

        # In case the simulation are very close to each other
        # or even collapse into a single point.
        if max_all - min_all < MIN_SIM_RANGE:
            average_value = (max_all + min_all) / 2
            delta_c = MIN_SIM_RANGE / (N - 1)
            if average_value < (MIN_SIM_RANGE / 2):
                c = np.linspace(0, MIN_SIM_RANGE, N)
            else:
                c = np.linspace(
                    average_value - (MIN_SIM_RANGE / 2),
                    average_value + (MIN_SIM_RANGE / 2),
                    N,
                )
            # Set the n(k) values for the simulations
            for i in range(len(sim_all)):
                n[i] = np.ceil((sim_all[i] - c[0]) / delta_c) + 1
                if n[i] > N:
                    n[i] = N
                    warnings.warn(
                        "Interval for a simulation has been set to a larger "
                        "value than the number of spline parameters.",
                        stacklevel=2,
                    )
        # In case the simulations are sufficiently apart:
        else:
            delta_c = (max_all - min_all) / (N - 1)
            c = np.linspace(min_all, max_all, N)
            for i in range(len(sim_all)):
                if i == max_idx:
                    n[i] = N
                elif i == min_idx:
                    n[i] = 1
                else:
                    n[i] = np.ceil((sim_all[i] - c[0]) / delta_c) + 1
                if n[i] > N:
                    n[i] = N

        n = n.astype(int)
        return delta_c, c, n

    def _get_minimal_difference(
        self,
        measurement_range: float,
        N: int,
        min_diff_factor: float,
    ):
        """Return minimal parameter difference for spline parameters."""
        return min_diff_factor * measurement_range / N

    def _get_inner_optimization_options(
        self,
        inner_parameters: list[SplineInnerParameter],
        N: int,
        min_meas: float,
        max_meas: float,
        min_diff: float,
    ) -> dict:
        """Return default options for scipy optimizer.

        Returns inner subproblem optimization options including startpoint
        and optimization bounds or constraints, dependent on solver method.

        Parameters
        ----------
        inner_parameters:
            Inner parameters of the spline group.
        N:
            Number of spline parameters.
        min_meas:
            Minimal measurement value.
        max_meas:
            Maximal measurement value.
        min_diff:
            Minimal difference between spline parameters.
        """
        range_all = max_meas - min_meas

        constraint_min_diff = np.full(N, min_diff)
        constraint_min_diff[0] = 0

        last_opt_values = np.asarray([x.value for x in inner_parameters])

        if (last_opt_values > 0).any():
            x0 = last_opt_values
        # In case this is the first inner optimization, initialize the
        # spline parameters to a linear function with a symmetric 60%
        # larger range than the measurement range.
        else:
            x0 = np.full(
                N,
                (
                    max_meas
                    + 0.3 * range_all
                    - np.max([min_meas - 0.3 * range_all, 0])
                )
                / (N - 1),
            )
            x0[0] = np.max([min_meas - 0.3 * range_all, 0])

        from scipy.optimize import Bounds

        inner_options = {
            "x0": x0,
            "method": "L-BFGS-B",
            "options": {"disp": None},
            "bounds": Bounds(lb=constraint_min_diff),
        }

        return inner_options


def _calculate_nllh_for_group(
    s: np.ndarray,
    sim_all: np.ndarray,
    measurements: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
    regularization_factor: float,
    regularize_spline: bool,
    group_dict: dict,
) -> float:
    """Calculate the negative log-likelihood for the group.

    Combines the sum of squared residuals, the noise parameter,
    and the regularization term to the negative log-likelihood.

    Parameters
    ----------
    s:
        Reformulated inner spline parameters.
    sim_all:
        Simulations for the group.
    measurements:
        Measurements for the group.
    N:
        Number of spline bases.
    delta_c:
        Distance between two spline bases.
    c:
        Spline bases.
    n:
        Indices of the spline bases.
    regularization_factor:
        Regularization factor.
    regularize_spline:
        Whether to regularize the spline.
    group_dict:
        Dictionary containing the group information.

    Returns
    -------
    Negative log-likelihood.
    """
    # Calculate residuals
    residuals_squared = _calculate_residuals_for_group(
        s=s,
        sim_all=sim_all,
        measurements=measurements,
        N=N,
        delta_c=delta_c,
        c=c,
        n=n,
    )
    K = len(sim_all)

    # Calculate sigma
    if group_dict[OPTIMIZE_NOISE]:
        sigma = _calculate_sigma_for_group(
            residuals_squared=residuals_squared,
            n_datapoints=N,
        )
        group_dict[INNER_NOISE_PARS] = sigma
    else:
        sigma = group_dict[INNER_NOISE_PARS]

    # Calculate regularization term
    if regularize_spline:
        regularization_term = _calculate_regularization_for_group(
            s=s,
            N=N,
            c=c,
            regularization_factor=regularization_factor,
        )
    else:
        regularization_term = 0.0

    # Combine all terms into the negative log-likelihood
    nllh = (
        0.5 * np.log(2 * np.pi * sigma**2) * K
        + residuals_squared / (sigma**2)
        + regularization_term
    )
    return nllh


def _calculate_nllh_gradient_for_group(
    s: np.ndarray,
    sim_all: np.ndarray,
    measurements: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
    regularization_factor: float,
    regularize_spline: bool,
    group_dict: dict,
) -> np.ndarray:
    """Calculate the gradient of the nllh wrt. spline differences s for the group.

    Combines the gradient of the sum of squared residuals and the gradient of the
    regularization term to the gradient of the negative log-likelihood.

    Parameters
    ----------
    s:
        Reformulated inner spline parameters.
    sim_all:
        Simulations for the group.
    measurements:
        Measurements for the group.
    N:
        Number of spline bases.
    delta_c:
        Distance between two spline bases.
    c:
        Spline bases.
    n:
        Indices of the spline bases.
    regularization_factor:
        Regularization factor.
    regularize_spline:
        Whether to regularize the spline.
    group_dict:
        Dictionary containing the group information.

    Returns
    -------
    Gradient of the negative log-likelihood wrt. spline differences s.
    """
    # Calculate gradient of residuals
    residuals_squared_gradient = _calculate_residuals_gradient_for_group(
        s=s,
        sim_all=sim_all,
        measurements=measurements,
        N=N,
        delta_c=delta_c,
        c=c,
        n=n,
    )

    # Calculate sigma
    if group_dict[OPTIMIZE_NOISE]:
        residuals_squared = _calculate_residuals_for_group(
            s=s,
            sim_all=sim_all,
            measurements=measurements,
            N=N,
            delta_c=delta_c,
            c=c,
            n=n,
        )
        sigma = _calculate_sigma_for_group(
            residuals_squared=residuals_squared,
            n_datapoints=N,
        )
        group_dict[INNER_NOISE_PARS] = sigma
    else:
        sigma = group_dict[INNER_NOISE_PARS]

    # Calculate gradient of regularization term
    if regularize_spline:
        regularization_term_gradient = (
            _calculate_regularization_gradient_for_group(
                s=s,
                N=N,
                c=c,
                regularization_factor=regularization_factor,
            )
        )
    else:
        regularization_term_gradient = np.zeros_like(s)

    # Combine all terms into the gradient of the negative log-likelihood
    nllh_gradient = (
        residuals_squared_gradient / (sigma**2) + regularization_term_gradient
    )
    return nllh_gradient


def _calculate_sigma_for_group(
    residuals_squared: float,
    n_datapoints: int,
):
    """Calculate the noise parameter sigma.

    Parameters
    ----------
    residuals_squared:
        The sum of squared residuals divided by 2.
    n_datapoints:
        The number of datapoints.
    """
    sigma = np.sqrt(2 * residuals_squared / n_datapoints)

    return sigma


def _calculate_residuals_for_group(
    s: np.ndarray,
    sim_all: np.ndarray,
    measurements: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Residuals squared for reformulated inner spline problem.

    Equal to 1/2 * sum_k (tilde{z}_k - z_k)^2
    """
    obj = 0

    for y_k, z_k, n_k in zip(sim_all, measurements, n, strict=True):
        i = n_k - 1
        sum_s = 0
        sum_s = np.sum(s[:i])
        if i == 0:
            obj += (z_k - s[i]) ** 2
        elif i == N:
            obj += (z_k - sum_s) ** 2
        else:
            obj += (z_k - (y_k - c[i - 1]) * s[i] / delta_c - sum_s) ** 2
    obj = obj / 2
    return obj


def _calculate_residuals_gradient_for_group(
    s: np.ndarray,
    sim_all: np.ndarray,
    measurements: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Gradient of the residuals with respect to the spline differences s_i for a group."""

    gradient = np.zeros(N)

    for y_k, z_k, n_k in zip(sim_all, measurements, n, strict=True):
        sum_s = 0
        i = n_k - 1  # just the iterator to go over the Jacobian array
        sum_s = np.sum(s[:i])
        if i == 0:
            gradient[i] += s[i] - z_k
        elif i == N:
            gradient[:i] += np.full(i, sum_s - z_k)
        else:
            gradient[i] += (
                ((y_k - c[i - 1]) * s[i] / delta_c + sum_s - z_k)
                * (y_k - c[i - 1])
                / delta_c
            )
            gradient[:i] += np.full(
                i,
                (y_k - c[i - 1]) * s[i] / delta_c + sum_s - z_k,
            )
    return gradient


def _calculate_regularization_for_group(
    s: np.ndarray,
    N: int,
    c: np.ndarray,
    regularization_factor: float,
):
    """Calculate regularization term the given spline.

    We regularize the spline to be linear. To do this, we calculate the optimal
    linear function that minimizes the sum of squared residuals with respect to
    the spline knots. Then we calculate the sum of squared residuals for this
    linear function. If the calculated offset is smaller than 0, we set it to 0.
    This is because the spline is not allowed to be negative.
    """
    # Calculate the spline knots xi_i from spline differences s_i
    lower_trian = np.tril(np.ones((N, N)))
    xi = np.dot(lower_trian, s)

    # Calculate auxiliary values
    c_sum = np.sum(c)
    xi_sum = np.sum(xi)
    c_squares_sum = np.sum(c**2)
    c_dot_xi = np.dot(c, xi)
    # Calculate the optimal linear function offset
    if np.isclose(N * c_squares_sum - c_sum**2, 0):
        beta_opt = xi_sum / N
    else:
        beta_opt = (xi_sum * c_squares_sum - c_dot_xi * c_sum) / (
            N * c_squares_sum - c_sum**2
        )

    # If the offset is smaller than 0, we set it to 0
    if beta_opt < 0:
        beta_opt = 0

    # Calculate the slope of the optimal linear function
    alpha_opt = (c_dot_xi - beta_opt * c_sum) / c_squares_sum

    # Calculate the sum of squared residuals for the optimal linear function
    regularization_term = np.sum((xi - alpha_opt * c - beta_opt) ** 2) / (
        2 * N
    )

    return regularization_term * regularization_factor


def _calculate_regularization_gradient_for_group(
    s: np.ndarray,
    N: int,
    c: np.ndarray,
    regularization_factor: float,
):
    """Calculate regularization term gradient for the given spline."""
    # Calculate the spline knots xi_i from spline differences s_i

    lower_trian = np.tril(np.ones((N, N)))
    xi = np.dot(lower_trian, s)

    # Calculate auxiliary values
    c_sum = np.sum(c)
    xi_sum = np.sum(xi)
    c_squares_sum = np.sum(c**2)
    c_dot_xi = np.dot(c, xi)

    # Calculate the optimal linear function offset
    if np.isclose(N * c_squares_sum - c_sum**2, 0):
        beta_opt = xi_sum / N
    else:
        beta_opt = (xi_sum * c_squares_sum - c_dot_xi * c_sum) / (
            N * c_squares_sum - c_sum**2
        )

    # If the offset is smaller than 0, we set it to 0.
    # Otherwise, we calculate the gradient of the offset.
    if beta_opt < 0:
        beta_opt = 0

    # Calculate the slope of the optimal linear function
    alpha_opt = (c_dot_xi - beta_opt * c_sum) / c_squares_sum

    # Calculate some more auxiliary values
    residuals = xi - alpha_opt * c - beta_opt

    # Can remove terms from this aux_matrix due to optimality
    # of the linear function (alpha & beta)
    aux_matrix = lower_trian

    # Calculate the gradient of the sum of squared residuals
    regularization_gradient = residuals @ aux_matrix / N

    return regularization_gradient * regularization_factor


def get_spline_mapped_simulations(
    s: np.ndarray,
    sim_all: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Return model simulations mapped using the approximation spline."""
    mapped_simulations = np.zeros(len(sim_all))
    lower_trian = np.tril(np.ones((N, N)))
    xi = np.dot(lower_trian, s)

    for y_k, n_k, index in zip(sim_all, n, range(len(sim_all)), strict=True):
        interval_index = n_k - 1
        if interval_index == 0 or interval_index == N:
            mapped_simulations[index] = xi[interval_index]
        else:
            mapped_simulations[index] = (y_k - c[interval_index - 1]) * (
                xi[interval_index] - xi[interval_index - 1]
            ) / delta_c + xi[interval_index - 1]

    return mapped_simulations


def calculate_inner_hessian(
    s: np.ndarray,
    sim_all: np.ndarray,
    sigma: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Calculate the hessian of the objective function for the reformulated inner problem."""

    hessian = np.zeros((N, N))

    for y_k, sigma_k, n_k in zip(sim_all, sigma, n, strict=True):
        sum_s = 0
        i = n_k - 1  # just the iterator to go over the Hessian matrix
        for j in range(i):
            sum_s += s[j]

        hessian[i][i] += (1 / sigma_k**2) * ((y_k - c[i - 1]) / delta_c) ** 2
        for j in range(i):
            hessian[i][j] += (1 / sigma_k**2) * ((y_k - c[i - 1]) / delta_c)
            hessian[j][i] += (1 / sigma_k**2) * ((y_k - c[i - 1]) / delta_c)
            for h in range(i):
                hessian[j][h] += 1 / sigma_k**2

    return hessian


def calculate_dy_term(
    sim_all: np.ndarray,
    sy_all: np.ndarray,
    measurements: np.ndarray,
    s: np.ndarray,
    N: int,
    delta_c: float,
    delta_c_dot: float,
    c: np.ndarray,
    c_dot: np.ndarray,
    n: np.ndarray,
):
    """Calculate the derivative of the objective function for one group with respect to the simulations."""
    df_dy = 0

    for y_k, z_k, y_dot_k, n_k in zip(
        sim_all, measurements, sy_all, n, strict=True
    ):
        i = n_k - 1
        sum_s = np.sum(s[:i])
        if i > 0 and i < N:
            df_dy += (
                ((y_k - c[i - 1]) * s[i] / delta_c + sum_s - z_k)
                * s[i]
                * (
                    (y_dot_k - c_dot[i - 1]) * delta_c
                    - (y_k - c[i - 1]) * delta_c_dot
                )
                / delta_c**2
            )
        # There is no i==0 case, because in this case
        # c[0] == y_k and so the derivative is zero.
    return df_dy


def calculate_spline_bases_gradient(
    sim_all: np.ndarray, sy_all: np.ndarray, N: int
):
    """Calculate gradient of the rescaled spline bases."""

    min_idx = np.argmin(sim_all)
    max_idx = np.argmax(sim_all)

    min_all = sim_all[min_idx]
    max_all = sim_all[max_idx]
    # Coming directly from differentiating _rescale_spline_bases
    if sim_all[max_idx] - sim_all[min_idx] < MIN_SIM_RANGE:
        delta_c_dot = 0
        c_dot = np.full(N, (sy_all[max_idx] - sy_all[min_idx]) / 2)
        average_value = (max_all + min_all) / 2
        if average_value < (MIN_SIM_RANGE / 2):
            c_dot = np.full(N, 0)
        else:
            c_dot = np.full(N, (sy_all[max_idx] - sy_all[min_idx]) / 2)
    else:
        delta_c_dot = (sy_all[max_idx] - sy_all[min_idx]) / (N - 1)
        c_dot = np.linspace(sy_all[min_idx], sy_all[max_idx], N)

    return delta_c_dot, c_dot


def extract_expdata_using_mask(
    expdata: list[np.ndarray], mask: list[np.ndarray]
):
    """Extract data from expdata list of arrays for the given mask."""
    return np.concatenate(
        [
            expdata[condition_index][mask[condition_index]]
            for condition_index in range(len(mask))
        ]
    )


def save_inner_parameters_to_inner_problem(
    inner_problem: SemiquantProblem,
    s: np.ndarray,
    group: int,
) -> None:
    """Save inner parameter values to the inner subproblem.

    Calculates the non-reformulated inner spline parameters from
    the reformulated inner spline parameters and saves them to
    the inner subproblem.

    Parameters
    ----------
    inner_parameters : list
        List of inner parameters.
    s : np.ndarray
        Reformulated inner spline parameters.
    """
    group_dict = inner_problem.groups[group]
    inner_spline_parameters = inner_problem.get_xs_for_group(group)
    inner_noise_parameters = inner_problem.get_noise_parameters_for_group(
        group
    )

    for idx in range(len(inner_spline_parameters)):
        inner_spline_parameters[idx].value = s[idx]

    sigma = group_dict[INNER_NOISE_PARS]

    if group_dict[OPTIMIZE_NOISE]:
        inner_noise_parameters[0].value = sigma


def get_monotonicity_measure(measurement, simulation):
    """Get monotonicity measure by calculating inversions.

    Calculates the number of inversions in the simulation data
    with respect to the measurement data.

    Parameters
    ----------
    measurement : np.ndarray
        Measurement data.
    simulation : np.ndarray
        Simulation data.

    Returns
    -------
    inversions : int
        Number of inversions.
    """
    if len(measurement) != len(simulation):
        raise ValueError(
            "Measurement and simulation data must have the same length."
        )

    ordered_simulation = [
        x
        for _, x in sorted(
            zip(measurement, simulation, strict=True), key=lambda pair: pair[0]
        )
    ]
    ordered_measurement = sorted(simulation)

    inversions = 0
    for i in range(len(ordered_simulation)):
        for j in range(i + 1, len(ordered_simulation)):
            if ordered_simulation[i] > ordered_simulation[j]:
                inversions += 1
            elif (
                ordered_simulation[i] == ordered_simulation[j]
                and ordered_measurement[i] != ordered_measurement[j]
            ):
                inversions += 1

    return inversions



"""Deprecated. Use ``petab_select.plot`` methods instead."""

import warnings

import matplotlib.pyplot as plt
import petab_select.plot
from petab_select import Criterion, Model, Models

from .. import select as pypesto_select


def default_label_maker(model: Model) -> str:
    """Create a model label, for plotting."""
    return str(model.hash)[:4]


# FIXME supply the problem to automatically detect the correct criterion?
def plot_selected_models(
    selected_models: list[Model],
    criterion: str = Criterion.AIC,
    relative: bool = True,
    fz: int = 14,
    size: tuple[float, float] = (5, 4),
    labels: dict[str, str] = None,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Use `petab_select.plot.line_best_by_iteration`` instead. Deprecated."""
    warnings.warn(
        "Use `petab_select.plot.line_best_by_iteration` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return petab_select.plot.line_best_by_iteration(
        models=Models(selected_models),
        criterion=criterion,
        relative=relative,
        fz=fz,
        labels=labels,
        ax=ax,
    )


def plot_calibrated_models_digraph(
    problem: pypesto_select.Problem,
    calibrated_models: dict[str, Model] = None,
    criterion: Criterion = None,
    optimal_distance: float = 1,
    relative: bool = True,
    options: dict = None,
    labels: dict[str, str] = None,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Use `petab_select.plot.graph_history`` instead. Deprecated."""
    warnings.warn(
        "Use `petab_select.plot.graph_history` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return petab_select.plot.graph_history(
        models=calibrated_models.values(),
        criterion=criterion,
        draw_networkx_kwargs=options,
        relative=relative,
        labels=labels,
        ax=ax,
    )



import warnings
from collections.abc import Sequence

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

import pypesto

from ..C import (
    AMICI_SIGMAY,
    AMICI_Y,
    CURRENT_SIMULATION,
    DATAPOINTS,
    EXPDATA_MASK,
    INNER_PARAMETERS,
    REGULARIZE_SPLINE,
    SCIPY_X,
    SPLINE_KNOTS,
    InnerParameterType,
)
from ..problem import HierarchicalProblem, Problem
from ..result import Result

try:
    import amici.sim.sundials as asd
    from amici.sim.sundials.petab.v1 import fill_in_parameters

    from ..hierarchical import InnerCalculatorCollector
    from ..hierarchical.base_problem import scale_back_value_dict
    from ..hierarchical.relative.calculator import RelativeAmiciCalculator
    from ..hierarchical.relative.problem import RelativeInnerProblem
    from ..hierarchical.semiquantitative.calculator import SemiquantCalculator
    from ..hierarchical.semiquantitative.solver import (
        SemiquantInnerSolver,
        _calculate_regularization_for_group,
        extract_expdata_using_mask,
        get_spline_mapped_simulations,
    )
except ImportError:
    pass


def visualize_estimated_observable_mapping(
    pypesto_result: Result,
    pypesto_problem: HierarchicalProblem,
    start_index: int = 0,
    axes: plt.Axes | None = None,
    **kwargs,
):
    """Visualize the estimated observable mapping for relative and semi-quantitative observables.

    Visualizes the estimated linear mapping for relative observables and the non-linear
    spline approximation for semi-quantitative observables.

    Parameters
    ----------
    pypesto_result:
        The pyPESTO result object from optimization.
    pypesto_problem:
        The pyPESTO problem. It should contain the objective object that was used for estimation.
    start_index:
        The observable mapping from this start's optimized vector will be plotted.
    axes:
        The axes to plot the estimated observable mapping on.
    kwargs:
        Additional arguments to passed to ``matplotlib.pyplot.subplots``
        (e.g. `figsize= ...`).

    Returns
    -------
    axes:
        The matplotlib axes.
    """

    # Check if the pyPESTO problem is hierarchical.
    if not isinstance(pypesto_problem, HierarchicalProblem):
        raise ValueError(
            "Only hierarchical problems contain estimated observable mappings. Please provide a hierarchical problem."
        )

    # Check if the pyPESTO problem contains an objective.
    if pypesto_problem.objective is None:
        raise ValueError(
            "The problem must contain the corresponding objective that was used for estimation."
        )

    # Check the calculator is the InnerCalculatorCollector.
    if not isinstance(
        pypesto_problem.objective.calculator, InnerCalculatorCollector
    ):
        raise ValueError(
            "The calculator must be an instance of the InnerCalculatorCollector."
        )

    amici_model = pypesto_problem.objective.amici_model

    # Get the number of relative and semi-quantitative observables.
    relative_obs_ids = pypesto_problem.relative_observable_ids or []
    semiquant_obs_ids = pypesto_problem.semiquant_observable_ids or []
    n_relative_observables = len(relative_obs_ids)
    n_semiquant_observables = len(semiquant_obs_ids)

    # Check if there are any relative or semi-quantitative observables.
    if n_relative_observables == 0 and n_semiquant_observables == 0:
        raise ValueError(
            "The problem does not contain any relative or semi-quantitative observables."
        )

    # Get observable indices for both relative and semi-quantitative observables.
    rel_and_semiquant_obs_indices = [
        amici_model.get_observable_ids().index(observable_id)
        for observable_id in relative_obs_ids + semiquant_obs_ids
    ]
    rel_and_semiquant_obs_indices.sort()

    # If axes are given, check if they are of the correct length.
    if (
        axes is not None
        and len(axes) != n_relative_observables + n_semiquant_observables
    ):
        raise ValueError(
            "The number of axes must be equal to the number of relative and semi-quantitative observables."
        )

    # If axes are not given, create them.
    if axes is None:
        n_axes = n_relative_observables + n_semiquant_observables
        n_rows = int(np.ceil(np.sqrt(n_axes)))
        n_cols = int(np.ceil(n_axes / n_rows))
        _, axes = plt.subplots(n_rows, n_cols, squeeze=False, **kwargs)
        axes = axes.flatten()

    # Plot the estimated observable mapping for relative observables.
    if n_relative_observables > 0:
        axes = plot_linear_observable_mappings_from_pypesto_result(
            pypesto_result=pypesto_result,
            pypesto_problem=pypesto_problem,
            start_index=start_index,
            axes=axes,
            rel_and_semiquant_obs_indices=rel_and_semiquant_obs_indices,
        )

    # Plot the estimated spline approximations for semi-quantitative observables.
    if n_semiquant_observables > 0:
        axes = plot_splines_from_pypesto_result(
            pypesto_result=pypesto_result,
            start_index=start_index,
            axes=axes,
            rel_and_semiquant_obs_indices=rel_and_semiquant_obs_indices,
        )

    # Remove any axes that were not used.
    for ax in axes[n_relative_observables + n_semiquant_observables :]:
        ax.remove()

    # Increase the distance between the subplots.
    plt.tight_layout()

    return axes


def plot_linear_observable_mappings_from_pypesto_result(
    pypesto_result: Result,
    pypesto_problem: HierarchicalProblem,
    start_index=0,
    axes: plt.Axes | None = None,
    rel_and_semiquant_obs_indices: list[int] | None = None,
    **kwargs,
):
    """Plot the linear observable mappings from a pyPESTO result.

    Parameters
    ----------
    pypesto_result:
        The pyPESTO result object from optimization.
    pypesto_problem:
        The pyPESTO problem. It should contain the objective object that was used for estimation.
    start_index:
        The observable mapping from this start's optimized vector will be plotted.
    axes:
        The axes to plot the linear observable mappings on.
    rel_and_semiquant_obs_indices:
        The indices of the relative and semi-quantitative observables in the
        amici model. Important if both relative and semi-quantitative observables
        will be plotted on the same axes.
    **kwargs:
        Additional arguments to pass to the ``matplotlib.pyplot.subplots`` function.

    Returns
    -------
    axes:
        The matplotlib axes.
    """
    # Check the calculator is the InnerCalculatorCollector.
    if not isinstance(
        pypesto_problem.objective.calculator, InnerCalculatorCollector
    ):
        raise ValueError(
            "The calculator must be an instance of the InnerCalculatorCollector."
        )

    # Get the needed objects from the pypesto problem.
    edatas = pypesto_problem.objective.edatas
    parameter_mapping = pypesto_problem.objective.parameter_mapping
    amici_model = pypesto_problem.objective.amici_model
    amici_solver = pypesto_problem.objective.amici_solver
    n_threads = pypesto_problem.objective.n_threads

    # Get the relative calculator.
    relative_calculator = [
        inner_calculator
        for inner_calculator in pypesto_problem.objective.calculator.inner_calculators
        if isinstance(inner_calculator, RelativeAmiciCalculator)
    ][0]

    # Get the inner problem and solver from the relative calculator.
    inner_problem: RelativeInnerProblem = relative_calculator.inner_problem

    # Get the relative observable ids and indices.
    relative_observable_ids = pypesto_problem.relative_observable_ids
    relative_observable_indices = [
        amici_model.get_observable_ids().index(observable_id)
        for observable_id in relative_observable_ids
    ]

    # Get the number of relative observables.
    n_relative_observables = len(relative_observable_ids)

    # Check if the axes are given.
    if axes is not None and len(axes) <= max(relative_observable_indices):
        raise ValueError(
            "The number of axes must be larger than the largest observable index."
        )
    # If axes are not given, create them.
    if axes is None:
        if n_relative_observables == 1:
            # Make figure with only one plot
            _, ax = plt.subplots(1, 1, **kwargs)

            axes = [ax]
        else:
            # Choose number of rows and columns to be used for the subplots
            n_rows = int(np.ceil(np.sqrt(n_relative_observables)))
            n_cols = int(np.ceil(n_relative_observables / n_rows))

            # Make as many subplots as there are relative observables
            _, axes = plt.subplots(n_rows, n_cols, squeeze=False, **kwargs)
            # Flatten the axes array
            axes = axes.flatten()

    #################################################################
    # Simulate the model with the parameters from the pypesto result.
    #################################################################

    # Get the parameters from the pypesto result for the start_index.
    x_dct = dict(
        zip(
            pypesto_problem.objective.x_ids,
            pypesto_result.optimize_result.list[start_index]["x"],
            strict=True,
        )
    )

    x_dct.update(
        pypesto_result.problem.objective.calculator.necessary_par_dummy_values
    )

    # Fill in the parameters.
    fill_in_parameters(
        edatas=edatas,
        problem_parameters=x_dct,
        scaled_parameters=True,
        parameter_mapping=parameter_mapping,
        amici_model=amici_model,
    )

    # Simulate the model with the parameters from the pypesto result.
    inner_rdatas = asd.run_simulations(
        amici_model,
        amici_solver,
        edatas,
        num_threads=min(n_threads, len(edatas)),
    )

    # If any amici simulation failed, raise warning and return None.
    if any(rdata.status != asd.AMICI_SUCCESS for rdata in inner_rdatas):
        raise ValueError(
            "Warning: Some AMICI simulations failed. Cannot plot inner "
            "solutions."
        )

    # Get the simulation.
    sim = [rdata[AMICI_Y] for rdata in inner_rdatas]

    # Get the inner parameters from the pypesto result.
    inner_parameter_values = dict(
        zip(
            pypesto_problem.inner_x_names,
            pypesto_result.optimize_result.list[start_index][INNER_PARAMETERS],
            strict=True,
        )
    )

    # Remove inner parameters not belonging to the relative inner problem.
    inner_parameter_values = {
        key: value
        for key, value in inner_parameter_values.items()
        if key in inner_problem.get_x_ids()
    }

    # Scale the inner parameters back to linear scale.
    inner_parameter_values = scale_back_value_dict(
        inner_parameter_values, inner_problem
    )

    ######################################
    # Plot the linear observable mappings.
    ######################################

    # plot the linear observable mapping for each relative_observable_id
    for observable_index, observable_id in zip(
        relative_observable_indices, relative_observable_ids, strict=True
    ):
        # Get the ax for the current observable.
        if rel_and_semiquant_obs_indices is not None:
            ax_index = rel_and_semiquant_obs_indices.index(observable_index)
        else:
            ax_index = relative_observable_indices.index(observable_index)

        ax = axes[ax_index]

        # Get the inner parameters for the current observable.
        inner_parameters = inner_problem.get_xs_for_obs_idx(observable_index)

        scaling_factor = None
        offset = None

        for inner_par in inner_parameters:
            if inner_par.inner_parameter_type == InnerParameterType.SCALING:
                scaling_factor = inner_par
            elif inner_par.inner_parameter_type == InnerParameterType.OFFSET:
                offset = inner_par

        scaling_factor_value = (
            inner_parameter_values[scaling_factor.inner_parameter_id]
            if scaling_factor is not None
            else 1
        )
        offset_value = (
            inner_parameter_values[offset.inner_parameter_id]
            if offset is not None
            else 0
        )

        # Get the data mask for the current observable.
        observable_data_mask = scaling_factor.ixs or offset.ixs

        # Get the measurements for the current observable.
        measurements = extract_expdata_using_mask(
            expdata=inner_problem.data, mask=observable_data_mask
        )

        # Get the simulation for the current observable.
        simulation = extract_expdata_using_mask(
            expdata=sim, mask=observable_data_mask
        )

        ax.plot(simulation, measurements, "bs", label="Measurements")

        # Plot the linear mapping.
        ax.plot(
            np.sort(simulation),
            scaling_factor_value * np.sort(simulation) + offset_value,
            linestyle="-",
            color="orange",
            label="Linear mapping",
        )

        ax.legend()
        ax.set_title(f"Observable {observable_id}")
        ax.set_xlabel("Model output")
        ax.set_ylabel("Measurements")

    if rel_and_semiquant_obs_indices is None:
        for ax in axes[n_relative_observables:]:
            ax.remove()

    return axes


def plot_splines_from_pypesto_result(
    pypesto_result: Result, start_index=0, **kwargs
):
    """Plot the estimated spline approximations from a pypesto result.

    Parameters
    ----------
    pypesto_result:
        The pypesto result.
    start_index:
        The observable mapping from this start's optimized vector will be plotted.
    kwargs:
        Additional arguments to pass to the plotting function.

    Returns
    -------
    axes:
        The matplotlib axes.
    """
    # Check that the problem contains an objective.
    if pypesto_result.problem.objective is None:
        raise ValueError(
            "The problem must contain the corresponding objective that was used for estimation."
        )

    # Check the calculator is the InnerCalculatorCollector.
    if not isinstance(
        pypesto_result.problem.objective.calculator, InnerCalculatorCollector
    ):
        raise ValueError(
            "The calculator must be an instance of the InnerCalculatorCollector."
        )

    # Check the result for start index contains the spline knots.
    if SPLINE_KNOTS not in pypesto_result.optimize_result.list[start_index]:
        raise ValueError(
            f"The result with index {start_index} does not contain the spline knots."
        )

    # Get the spline knot values from the pypesto result
    spline_knot_values = [
        obs_spline_knots[1]
        for obs_spline_knots in pypesto_result.optimize_result.list[
            start_index
        ][SPLINE_KNOTS]
    ]

    # Get inner parameters per observable as differences of spline knot values
    inner_parameters = [
        np.concatenate([[obs_knot_values[0]], np.diff(obs_knot_values)])
        for obs_knot_values in spline_knot_values
    ]

    inner_results = [
        {SCIPY_X: obs_inner_parameter}
        for obs_inner_parameter in inner_parameters
    ]

    # Get the parameters from the pypesto result for the start_index.
    x_dct = dict(
        zip(
            pypesto_result.problem.objective.x_ids,
            pypesto_result.optimize_result.list[start_index]["x"],
            strict=True,
        )
    )

    x_dct.update(
        pypesto_result.problem.objective.calculator.necessary_par_dummy_values
    )

    # Get the needed objects from the pypesto problem.
    edatas = pypesto_result.problem.objective.edatas
    parameter_mapping = pypesto_result.problem.objective.parameter_mapping
    amici_model = pypesto_result.problem.objective.amici_model
    amici_solver = pypesto_result.problem.objective.amici_solver
    n_threads = pypesto_result.problem.objective.n_threads
    observable_ids = amici_model.get_observable_ids()

    # Fill in the parameters.
    fill_in_parameters(
        edatas=edatas,
        problem_parameters=x_dct,
        scaled_parameters=True,
        parameter_mapping=parameter_mapping,
        amici_model=amici_model,
    )

    # Simulate the model with the parameters from the pypesto result.
    inner_rdatas = asd.run_simulations(
        amici_model,
        amici_solver,
        edatas,
        num_threads=min(n_threads, len(edatas)),
    )

    # If any amici simulation failed, raise warning and return None.
    if any(rdata.status != asd.AMICI_SUCCESS for rdata in inner_rdatas):
        warnings.warn(
            "Warning: Some AMICI simulations failed. Cannot plot inner "
            "solutions.",
            stacklevel=2,
        )
        return None

    # Get simulation.
    sim = [rdata[AMICI_Y] for rdata in inner_rdatas]

    spline_calculator = None
    for (
        calculator
    ) in pypesto_result.problem.objective.calculator.inner_calculators:
        if isinstance(calculator, SemiquantCalculator):
            spline_calculator = calculator
            break

    if spline_calculator is None:
        raise ValueError(
            "No SemiquantCalculator found in the inner_calculators of the objective. "
            "Cannot plot splines."
        )

    # Get the inner solver and problem.
    inner_solver = spline_calculator.inner_solver
    inner_problem = spline_calculator.inner_problem

    return plot_splines_from_inner_result(
        inner_problem,
        inner_solver,
        inner_results,
        sim,
        observable_ids,
        **kwargs,
    )


def plot_splines_from_inner_result(
    inner_problem: "pypesto.hierarchical.spline_approximation.problem.SplineInnerProblem",
    inner_solver: "pypesto.hierarchical.spline_approximation.solver.SplineInnerSolver",
    results: list[dict],
    sim: list[np.ndarray],
    observable_ids=None,
    axes: plt.Axes | None = None,
    rel_and_semiquant_obs_indices: list[int] | None = None,
    **kwargs,
):
    """Plot the estimated spline approximations from inner results.

    Parameters
    ----------
    inner_problem:
        The inner problem.
    inner_solver:
        The inner solver.
    results:
        The results from the inner solver.
    sim:
        The simulated model output.
    observable_ids:
        The ids of the observables.
    axes:
        The axes to plot the estimated spline approximations on.
    rel_and_semiquant_obs_indices:
        The indices of the relative and semi-quantitative observables in the
        amici model. Important if both relative and semi-quantitative observables
        will be plotted on the same axes.
    kwargs:
        Additional arguments to pass to the plotting function.

    Returns
    -------
    axes:
        The matplotlib axes.
    """

    if len(results) != len(inner_problem.groups):
        raise ValueError(
            "Number of results must be equal to number of groups of the inner subproblem."
        )

    # Get the number of groups
    n_groups = len(inner_problem.groups)
    semiquant_groups = list(inner_problem.groups.keys())

    # Check if the axes are given
    if axes is not None and len(axes) < max(semiquant_groups):
        raise ValueError(
            "The number of axes must be equal to or larger than the largest group index."
        )

    if axes is None:
        if n_groups == 1:
            # Make figure with only one plot
            _, ax = plt.subplots(1, 1, **kwargs)

            axes = [ax]
        else:
            # Choose number of rows and columns to be used for the subplots
            n_rows = int(np.ceil(np.sqrt(n_groups)))
            n_cols = int(np.ceil(n_groups / n_rows))

            # Make as many subplots as there are groups
            _, axes = plt.subplots(n_rows, n_cols, squeeze=False, **kwargs)
            # Flatten the axes array
            axes = axes.flatten()

    # for each result and group, plot the inner solution
    for result, group in zip(results, inner_problem.groups, strict=True):
        if rel_and_semiquant_obs_indices is not None:
            ax_index = rel_and_semiquant_obs_indices.index(group - 1)
        else:
            ax_index = semiquant_groups.index(group)

        # For each group get the inner parameters and simulation
        s = result[SCIPY_X]

        # Utility matrix for the spline knot calculation
        lower_trian = np.tril(np.ones((len(s), len(s))))
        spline_knots = np.dot(lower_trian, s)

        measurements = inner_problem.groups[group][DATAPOINTS]
        simulation = extract_expdata_using_mask(
            expdata=sim, mask=inner_problem.groups[group][EXPDATA_MASK]
        )

        # For the simulation, get the spline bases
        (
            delta_c,
            spline_bases,
            n,
        ) = SemiquantInnerSolver._rescale_spline_bases(
            sim_all=simulation,
            N=len(spline_knots),
            K=len(simulation),
        )

        axes[ax_index].plot(
            simulation, measurements, "bs", label="Measurements"
        )
        axes[ax_index].plot(
            spline_bases, spline_knots, "g.", label="Spline knots"
        )
        axes[ax_index].plot(
            spline_bases,
            spline_knots,
            linestyle="-",
            color="g",
            label="Spline function",
        )
        if inner_solver.options[REGULARIZE_SPLINE]:
            alpha_opt, beta_opt = _calculate_optimal_regularization(
                s=s,
                N=len(spline_knots),
                c=spline_bases,
            )
            axes[ax_index].plot(
                spline_bases,
                alpha_opt * spline_bases + beta_opt,
                linestyle="--",
                color="orange",
                label="Regularization line",
            )

        axes[ax_index].legend()
        if observable_ids is not None:
            axes[ax_index].set_title(f"Observable {observable_ids[group - 1]}")
        else:
            axes[ax_index].set_title(f"Group {group}")

        axes[ax_index].set_xlabel("Model output")
        axes[ax_index].set_ylabel("Measurements")

    if rel_and_semiquant_obs_indices is None:
        for ax in axes[len(results) :]:
            ax.remove()

    return axes


def _calculate_optimal_regularization(
    s: np.ndarray,
    N: int,
    c: np.ndarray,
):
    """Calculate the optimal linear regularization for the spline approximation.

    Parameters
    ----------
    s:
        The spline parameters.
    N:
        The number of inner parameters.
    c:
        The spline bases.

    Returns
    -------
    alpha_opt:
        The optimal slope of the linear function.
    beta_opt:
        The optimal offset of the linear function.
    """
    lower_trian = np.tril(np.ones((N, N)))
    xi = np.dot(lower_trian, s)

    # Calculate auxiliary values
    c_sum = np.sum(c)
    xi_sum = np.sum(xi)
    c_squares_sum = np.sum(c**2)
    c_dot_xi = np.dot(c, xi)
    # Calculate the optimal linear function offset
    if np.isclose(N * c_squares_sum - c_sum**2, 0):
        beta_opt = xi_sum / N
    else:
        beta_opt = (xi_sum * c_squares_sum - c_dot_xi * c_sum) / (
            N * c_squares_sum - c_sum**2
        )

    # If the offset is smaller than 0, we set it to 0
    if beta_opt < 0:
        beta_opt = 0

    # Calculate the slope of the optimal linear function
    alpha_opt = (c_dot_xi - beta_opt * c_sum) / c_squares_sum

    return alpha_opt, beta_opt


def _add_spline_mapped_simulations_to_model_fit(
    result: Result | Sequence[Result],
    pypesto_problem: Problem,
    start_index: int = 0,
    axes: plt.Axes | None = None,
) -> matplotlib.axes.Axes | None:
    """Visualize the spline optimized model fit.

    Adds the spline-mapped simulation to the axes given by
    `pypesto.visualize.model_fit.visualize_optimized_model_fit`.
    For further details on documentation see
    :py:func:`pypesto.visualize.model_fit.visualize_optimized_model_fit`.
    """

    # If no visualize_optimized_model_fit axes were given,
    # return None.
    if axes is None:
        return None

    # Get the parameters from the pypesto result for the start_index.
    x_dct = dict(
        zip(
            pypesto_problem.objective.x_ids,
            result.optimize_result.list[start_index]["x"],
            strict=True,
        )
    )
    x_dct.update(
        pypesto_problem.objective.calculator.necessary_par_dummy_values
    )
    # Get the needed objects from the pypesto problem.
    edatas = pypesto_problem.objective.edatas
    parameter_mapping = pypesto_problem.objective.parameter_mapping
    amici_model = pypesto_problem.objective.amici_model
    amici_solver = pypesto_problem.objective.amici_solver
    n_threads = pypesto_problem.objective.n_threads

    # Fill in the parameters.
    fill_in_parameters(
        edatas=edatas,
        problem_parameters=x_dct,
        scaled_parameters=True,
        parameter_mapping=parameter_mapping,
        amici_model=amici_model,
    )

    # Simulate the model with the parameters from the pypesto result.
    inner_rdatas = asd.run_simulations(
        amici_model,
        amici_solver,
        edatas,
        num_threads=min(n_threads, len(edatas)),
    )

    # If any amici simulation failed, raise warning and return None.
    if any(rdata.status != asd.AMICI_SUCCESS for rdata in inner_rdatas):
        warnings.warn(
            "Warning: Some AMICI simulations failed. Cannot plot inner "
            "solutions.",
            stacklevel=2,
        )
        return None

    # Get simulation and sigma.
    sim = [rdata[AMICI_Y] for rdata in inner_rdatas]
    sigma = [rdata[AMICI_SIGMAY] for rdata in inner_rdatas]

    spline_calculator = None
    for calculator in pypesto_problem.objective.calculator.inner_calculators:
        if isinstance(calculator, SemiquantCalculator):
            spline_calculator = calculator
            break

    # Get the inner solver and problem.
    inner_solver = spline_calculator.inner_solver
    inner_problem = spline_calculator.inner_problem

    # Solve the inner problem.
    inner_results = inner_solver.solve(inner_problem, sim, sigma)

    # Get the observable ids.
    observable_ids = amici_model.get_observable_ids()

    for inner_result, group in zip(
        inner_results, inner_problem.groups, strict=True
    ):
        observable_id = observable_ids[group - 1]
        # Get the ax for the current observable.
        ax = [
            ax
            for ax in axes.values()
            if observable_id in ax.legend().get_texts()[0].get_text().split()
        ][0]

        # Get the inner parameters and simulation.
        s = inner_result[SCIPY_X]
        simulation = inner_problem.groups[group][CURRENT_SIMULATION]

        # For the simulation, get the spline bases
        (
            delta_c,
            spline_bases,
            n,
        ) = SemiquantInnerSolver._rescale_spline_bases(
            sim_all=simulation,
            N=len(s),
            K=len(simulation),
        )
        # and the spline-mapped simulations.
        mapped_simulations = get_spline_mapped_simulations(
            s, simulation, len(s), delta_c, spline_bases, n
        )

        # Plot the spline-mapped simulations to the ax with same color
        # and timepoints as the lines which have 'simulation' in their label.
        plotted_index = 0
        for line in ax.lines:
            if "simulation" in line.get_label():
                color = line.get_color()
                timepoints = line.get_xdata()
                condition_mapped_simulations = mapped_simulations[
                    plotted_index : len(timepoints) + plotted_index
                ]
                plotted_index += len(timepoints)

                ax.plot(
                    timepoints,
                    condition_mapped_simulations,
                    color=color,
                    linestyle="dotted",
                    marker="^",
                )

        # Add linestyle='dotted' and marker='^' to the legend as black spline mapped simulations.
        ax.plot(
            [],
            [],
            color="black",
            linestyle="dotted",
            marker="^",
            label="Spline mapped simulation",
        )

    # Reset the legend.
    for ax in axes.values():
        ax.legend()

    return axes


def _obtain_regularization_for_start(
    pypesto_result: Result, start_index=0
) -> float | None:
    """Obtain the regularization for the start index.

    Calculates and returns the spline linear regularization
    term of the objective function for the start index.

    Parameters
    ----------
    pypesto_result:
        The pypesto result.
    start_index:
        The start index for which to calculate the regularization.

    Returns
    -------
    The regularization term of the objective function for the start index.
    """
    # Get the parameters from the pypesto result for the start_index.
    x_dct = dict(
        zip(
            pypesto_result.problem.objective.x_ids,
            pypesto_result.optimize_result.list[start_index]["x"],
            strict=True,
        )
    )

    x_dct.update(
        pypesto_result.problem.objective.calculator.necessary_par_dummy_values
    )

    # Get the needed objects from the pypesto problem.
    edatas = pypesto_result.problem.objective.edatas
    parameter_mapping = pypesto_result.problem.objective.parameter_mapping
    amici_model = pypesto_result.problem.objective.amici_model
    amici_solver = pypesto_result.problem.objective.amici_solver
    n_threads = pypesto_result.problem.objective.n_threads

    # Fill in the parameters.
    fill_in_parameters(
        edatas=edatas,
        problem_parameters=x_dct,
        scaled_parameters=True,
        parameter_mapping=parameter_mapping,
        amici_model=amici_model,
    )

    # Simulate the model with the parameters from the pypesto result.
    inner_rdatas = asd.run_simulations(
        amici_model,
        amici_solver,
        edatas,
        num_threads=min(n_threads, len(edatas)),
    )

    # If any amici simulation failed, raise warning and return None.
    if any(rdata.status != asd.AMICI_SUCCESS for rdata in inner_rdatas):
        warnings.warn(
            "Warning: Some AMICI simulations failed. Cannot plot inner "
            "solutions.",
            stacklevel=2,
        )
        return None

    # Get simulation and sigma.
    sim = [rdata[AMICI_Y] for rdata in inner_rdatas]
    sigma = [rdata[AMICI_SIGMAY] for rdata in inner_rdatas]

    spline_calculator = None
    for (
        calculator
    ) in pypesto_result.problem.objective.calculator.inner_calculators:
        if isinstance(calculator, SemiquantCalculator):
            spline_calculator = calculator
            break

    # Get the inner solver and problem.
    inner_solver = spline_calculator.inner_solver
    inner_problem = spline_calculator.inner_problem

    inner_results = inner_solver.solve(inner_problem, sim, sigma)

    reg_term_sum = 0

    # for each result and group, plot the inner solution
    for result, group in zip(inner_results, inner_problem.groups, strict=True):
        # For each group get the inner parameters and simulation
        s = result[SCIPY_X]
        simulation = inner_problem.groups[group][CURRENT_SIMULATION]

        # For the simulation, get the spline bases
        (
            _,
            spline_bases,
            _,
        ) = SemiquantInnerSolver._rescale_spline_bases(
            sim_all=simulation,
            N=len(s),
            K=len(simulation),
        )

        if inner_solver.options[REGULARIZE_SPLINE]:
            reg_term = _calculate_regularization_for_group(
                s=s,
                N=len(s),
                c=spline_bases,
                regularization_factor=inner_solver.options[
                    "regularization_factor"
                ],
            )
            reg_term_sum += reg_term

    return reg_term_sum



from __future__ import annotations

import copy
from collections.abc import Sequence

import numpy as np

try:
    import amici.sim.sundials as asd
    from amici.sim._parameter_mapping import ParameterMapping
    from amici.sim.sundials.petab.v1 import fill_in_parameters
except ImportError:
    pass

from ...C import (
    AMICI_SIGMAY,
    AMICI_SSIGMAY,
    AMICI_SY,
    AMICI_Y,
    FVAL,
    GRAD,
    HESS,
    INNER_PARAMETERS,
    RDATAS,
    RES,
    SRES,
    ModeType,
)
from ...objective.amici.amici_calculator import (
    AmiciCalculator,
    AmiciModel,
    AmiciSolver,
)
from ...objective.amici.amici_util import (
    filter_return_dict,
    init_return_values,
)
from .problem import AmiciInnerProblem
from .solver import AnalyticalInnerSolver, InnerSolver


class RelativeAmiciCalculator(AmiciCalculator):
    """A calculator that is passed as `calculator` to the pypesto.AmiciObjective."""

    def __init__(
        self,
        inner_problem: AmiciInnerProblem,
        inner_solver: InnerSolver | None = None,
    ):
        """Initialize the calculator from the given problem.

        Arguments
        ---------
        inner_problem:
            The inner problem of a hierarchical optimization problem.
        inner_solver:
            A solver to solve ``inner_problem``.
            Defaults to ``pypesto.hierarchical.solver.AnalyticalInnerSolver``.
        """
        super().__init__()

        self.inner_problem = inner_problem

        if inner_solver is None:
            inner_solver = AnalyticalInnerSolver()
        self.inner_solver = inner_solver

    def initialize(self):
        """Initialize."""
        super().initialize()
        self.inner_solver.initialize()

    def __call__(
        self,
        x_dct: dict,
        sensi_orders: tuple[int],
        mode: ModeType,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: list[asd.ExpData],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
        rdatas: list[asd.ReturnData] = None,
    ):
        """Perform the actual AMICI call, with hierarchical optimization.

        The return object also includes the simulation results that were
        generated to solve the inner problem, as well as the parameters that
        solver the inner problem.

        Parameters
        ----------
        x_dct:
            Parameters for which to compute function value and derivatives.
        sensi_orders:
            Tuple of requested sensitivity orders.
        mode:
            Call mode (function value or residual based).
        amici_model:
            The AMICI model.
        amici_solver:
            The AMICI solver.
        edatas:
            The experimental data.
        n_threads:
            Number of threads for AMICI call.
        x_ids:
            Ids of optimization parameters.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        fim_for_hess:
            Whether to use the FIM (if available) instead of the Hessian (if
            requested).
        rdatas:
            AMICI simulation return data. In case the calculator is part of
            the :class:`pypesto.objective.amici.InnerCalculatorCollector`,
            it will already simulate the model and pass the results here.

        Returns
        -------
        inner_result:
            A dict containing the calculation results: FVAL, GRAD, RDATAS and INNER_PARAMETERS.
        """
        if not self.inner_problem.check_edatas(edatas=edatas):
            raise ValueError(
                "The experimental data provided to this call differs from "
                "the experimental data used to setup the hierarchical "
                "optimizer."
            )

        if (
            1 in sensi_orders
            and amici_solver.get_sensitivity_method()
            == asd.SensitivityMethod.adjoint
        ) or 2 in sensi_orders:
            inner_result, inner_parameters = self.call_amici_twice(
                x_dct=x_dct,
                sensi_orders=sensi_orders,
                mode=mode,
                amici_model=amici_model,
                amici_solver=amici_solver,
                edatas=edatas,
                n_threads=n_threads,
                x_ids=x_ids,
                parameter_mapping=parameter_mapping,
                fim_for_hess=fim_for_hess,
            )
        else:
            inner_result, inner_parameters = self.calculate_directly(
                x_dct=x_dct,
                sensi_orders=sensi_orders,
                mode=mode,
                amici_model=amici_model,
                amici_solver=amici_solver,
                edatas=edatas,
                n_threads=n_threads,
                x_ids=x_ids,
                parameter_mapping=parameter_mapping,
                fim_for_hess=fim_for_hess,
                rdatas=rdatas,
            )

        inner_result[INNER_PARAMETERS] = (
            np.array(
                [
                    inner_parameters[x_id]
                    for x_id in self.inner_problem.get_x_ids()
                ]
            )
            if inner_parameters is not None
            else None
        )

        return inner_result

    def call_amici_twice(
        self,
        x_dct: dict,
        sensi_orders: tuple[int],
        mode: ModeType,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: list[asd.ExpData],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
    ):
        """Calculate by calling AMICI twice.

        This is necessary if the adjoint method is used, or if the Hessian is
        requested. In these cases, AMICI is called first to obtain simulations
        for the calculation of the inner parameters, and then again to obtain
        the requested objective function and gradient through AMICI.
        """
        dim = len(x_ids)
        # compute optimal inner parameters
        x_dct = copy.deepcopy(x_dct)
        x_dct.update(self.inner_problem.get_dummy_values(scaled=True))

        inner_result = super().__call__(
            x_dct=x_dct,
            sensi_orders=(0,),
            mode=mode,
            amici_model=amici_model,
            amici_solver=amici_solver,
            edatas=edatas,
            n_threads=n_threads,
            x_ids=x_ids,
            parameter_mapping=parameter_mapping,
            fim_for_hess=fim_for_hess,
        )
        rdatas = inner_result[RDATAS]

        # if any amici simulation failed, it's unlikely we can compute
        # meaningful inner parameters, so we better just fail early.
        if any(rdata.status != asd.AMICI_SUCCESS for rdata in rdatas):
            # if the gradient was requested, we need to provide some value
            # for it
            if 1 in sensi_orders:
                inner_result[GRAD] = np.full(shape=dim, fill_value=np.nan)
            if 2 in sensi_orders:
                inner_result[HESS] = np.full(
                    shape=(dim, dim), fill_value=np.nan
                )
            return inner_result, None

        inner_parameters = self.inner_solver.solve(
            problem=self.inner_problem,
            sim=[rdata[AMICI_Y] for rdata in rdatas],
            sigma=[rdata[AMICI_SIGMAY] for rdata in rdatas],
            scaled=True,
        )

        # fill in optimal values
        # directly writing to parameter mapping ensures that plists do not
        # include hierarchically computed parameters
        x_dct = copy.deepcopy(x_dct)
        x_dct.update(inner_parameters)

        # TODO use plist to compute only required derivatives, in
        #  `super.__call__`, `amici.parameter_mapping.fill_in_parameters`
        inner_result = super().__call__(
            x_dct=x_dct,
            sensi_orders=sensi_orders,
            mode=mode,
            amici_model=amici_model,
            amici_solver=amici_solver,
            edatas=edatas,
            n_threads=n_threads,
            x_ids=x_ids,
            parameter_mapping=parameter_mapping,
            fim_for_hess=fim_for_hess,
        )
        return inner_result, inner_parameters

    def calculate_directly(
        self,
        x_dct: dict,
        sensi_orders: tuple[int],
        mode: ModeType,
        amici_model: AmiciModel,
        amici_solver: AmiciSolver,
        edatas: list[asd.ExpData],
        n_threads: int,
        x_ids: Sequence[str],
        parameter_mapping: ParameterMapping,
        fim_for_hess: bool,
        rdatas: list[asd.ReturnData] = None,
    ):
        """Calculate directly via solver calculate methods.

        This is possible if the forward method is used, and the Hessian is not
        requested. In this case, the objective function and gradient are computed
        directly using the solver methods.
        """
        dim = len(x_ids)
        # compute optimal inner parameters
        x_dct = copy.deepcopy(x_dct)

        # initialize return values
        nllh, snllh, s2nllh, chi2, res, sres = init_return_values(
            sensi_orders, mode, dim
        )

        # set order in solver
        sensi_order = 0
        if sensi_orders:
            sensi_order = max(sensi_orders)

        # if AMICI ReturnData is not provided, we need to simulate the model
        if rdatas is None:
            amici_solver.set_sensitivity_order(sensi_order)
            x_dct.update(self.inner_problem.get_dummy_values(scaled=True))
            # fill in parameters
            fill_in_parameters(
                edatas=edatas,
                problem_parameters=x_dct,
                scaled_parameters=True,
                parameter_mapping=parameter_mapping,
                amici_model=amici_model,
            )
            # run amici simulation
            rdatas = asd.run_simulations(
                amici_model,
                amici_solver,
                edatas,
                num_threads=min(n_threads, len(edatas)),
            )

        inner_result = {
            FVAL: nllh,
            GRAD: snllh,
            HESS: s2nllh,
            RES: res,
            SRES: sres,
            RDATAS: rdatas,
        }

        # if any amici simulation failed, it's unlikely we can compute
        # meaningful inner parameters, so we better just fail early.
        if any(rdata.status != asd.AMICI_SUCCESS for rdata in rdatas):
            inner_result[FVAL] = np.inf
            if 1 in sensi_orders:
                inner_result[GRAD] = np.full(
                    shape=len(x_ids), fill_value=np.nan
                )
            return filter_return_dict(inner_result), None

        inner_parameters = self.inner_solver.solve(
            problem=self.inner_problem,
            sim=[rdata[AMICI_Y] for rdata in rdatas],
            sigma=[rdata[AMICI_SIGMAY] for rdata in rdatas],
            scaled=True,
        )

        # compute the objective function value
        inner_result[FVAL] = self.inner_solver.calculate_obj_function(
            problem=self.inner_problem,
            sim=[rdata[AMICI_Y] for rdata in rdatas],
            sigma=[rdata[AMICI_SIGMAY] for rdata in rdatas],
            inner_parameters=inner_parameters,
        )

        # compute the objective function gradient, if requested
        if 1 in sensi_orders:
            inner_result[GRAD] = self.inner_solver.calculate_gradients(
                problem=self.inner_problem,
                sim=[rdata[AMICI_Y] for rdata in rdatas],
                ssim=[rdata[AMICI_SY] for rdata in rdatas],
                sigma=[rdata[AMICI_SIGMAY] for rdata in rdatas],
                ssigma=[rdata[AMICI_SSIGMAY] for rdata in rdatas],
                inner_parameters=inner_parameters,
                parameter_mapping=parameter_mapping,
                par_opt_ids=x_ids,
                par_sim_ids=amici_model.get_free_parameter_ids(),
                snllh=snllh,
            )
        # apply the computed inner parameters to the ReturnData
        rdatas = self.inner_solver.apply_inner_parameters_to_rdatas(
            problem=self.inner_problem,
            rdatas=rdatas,
            inner_parameters=inner_parameters,
        )
        inner_result[RDATAS] = rdatas

        return inner_result, inner_parameters



"""Optimization result."""

import logging
import warnings
from collections import Counter
from collections.abc import Sequence
from copy import deepcopy
from typing import Union

import h5py
import numpy as np
import pandas as pd

from ..history import HistoryBase
from ..problem import Problem
from ..util import assign_clusters, delete_nan_inf

OptimizationResult = Union["OptimizerResult", "OptimizeResult"]
logger = logging.getLogger(__name__)


class OptimizerResult(dict):
    """
    The result of an optimizer run.

    Used as a standardized return value to map from the individual result
    objects returned by the employed optimizers to the format understood by
    pypesto.

    Can be used like a dict.

    Attributes
    ----------
    id:
        Id of the optimizer run. Usually the start index.
    x:
        The best found parameters.
    fval:
        The best found function value, `fun(x)`.
    grad:
        The gradient at `x`.
    hess:
        The Hessian at `x`.
    res:
        The residuals at `x`.
    sres:
        The residual sensitivities at `x`.
    n_fval
        Number of function evaluations.
    n_grad:
        Number of gradient evaluations.
    n_hess:
        Number of Hessian evaluations.
    n_res:
        Number of residuals evaluations.
    n_sres:
        Number of residual sensitivity evaluations.
    x0:
        The starting parameters.
    fval0:
        The starting function value, `fun(x0)`.
    history:
        Objective history.
    exitflag:
        The exitflag of the optimizer.
    time:
        Execution time.
    message: str
        Textual comment on the optimization result.
    optimizer: str
        The optimizer used for optimization.

    Notes
    -----
    Any field not supported by the optimizer is filled with None.
    """

    def __init__(
        self,
        id: str = None,
        x: np.ndarray = None,
        fval: float = None,
        grad: np.ndarray = None,
        hess: np.ndarray = None,
        res: np.ndarray = None,
        sres: np.ndarray = None,
        n_fval: int = None,
        n_grad: int = None,
        n_hess: int = None,
        n_res: int = None,
        n_sres: int = None,
        x0: np.ndarray = None,
        fval0: float = None,
        history: HistoryBase = None,
        exitflag: int = None,
        time: float = None,
        message: str = None,
        optimizer: str = None,
    ):
        super().__init__()
        self.id = id
        self.x: np.ndarray = np.array(x) if x is not None else None
        self.fval: float = fval
        self.grad: np.ndarray = np.array(grad) if grad is not None else None
        self.hess: np.ndarray = np.array(hess) if hess is not None else None
        self.res: np.ndarray = np.array(res) if res is not None else None
        self.sres: np.ndarray = np.array(sres) if sres is not None else None
        self.n_fval: int = n_fval
        self.n_grad: int = n_grad
        self.n_hess: int = n_hess
        self.n_res: int = n_res
        self.n_sres: int = n_sres
        self.x0: np.ndarray = np.array(x0) if x0 is not None else None
        self.fval0: float = fval0
        self.history: HistoryBase = history
        self.exitflag: int = exitflag
        self.time: float = time
        self.message: str = message
        self.optimizer = optimizer
        self.free_indices = None
        self.inner_parameters = None
        self.spline_knots = None

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key) from None

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def summary(self, full: bool = False, show_hess: bool = True) -> str:
        """
        Get summary of the object.

        Parameters
        ----------
        full:
            If True, print full vectors including fixed parameters.
        show_hess:
            If True, display the Hessian of the result.

        Returns
        -------
        summary: str
        """
        # add warning, if self.free_indices is None
        if self.free_indices is None:
            if full:
                logger.warning(
                    "There is no information about fixed parameters, "
                    "run update_to_full with the corresponding problem first."
                )
            full = True
        message = (
            "### Optimizer Result\n\n"
            f"* optimizer used: {self.optimizer}\n"
            f"* message: {self.message} \n"
            f"* number of evaluations: {self.n_fval}\n"
            f"* time taken to optimize: {self.time:0.3f}s\n"
            f"* startpoint: {self.x0 if full or self.x0 is None else self.x0[self.free_indices]}\n"
            f"* endpoint: {self.x if full else self.x[self.free_indices]}\n"
        )
        # add fval, gradient, hessian, res, sres if available
        if self.fval is not None:
            message += f"* final objective value: {self.fval}\n"
        if self.grad is not None:
            message += (
                f"* final gradient value: "
                f"{self.grad if full else self.grad[self.free_indices]}\n"
            )
        if self.hess is not None and show_hess:
            hess = self.hess
            if not full:
                hess = self.hess[np.ix_(self.free_indices, self.free_indices)]
            message += f"* final hessian value: {hess}\n"
        if self.res is not None:
            message += f"* final residual value: {self.res}\n"
        if self.sres is not None:
            message += f"* final residual sensitivity: {self.sres}\n"

        return message

    def update_to_full(self, problem: Problem) -> None:
        """
        Update values to full vectors/matrices.

        Parameters
        ----------
        problem:
            problem which contains info about how to convert to full vectors
            or matrices
        """
        self.x = problem.get_full_vector(self.x)
        self.grad = problem.get_full_vector(self.grad, x_is_grad=True)
        self.hess = problem.get_full_matrix(self.hess)
        self.x0 = problem.get_full_vector(self.x0)
        self.free_indices = np.array(problem.x_free_indices)


class OptimizeResult:
    """Result of the :py:func:`pypesto.optimize.minimize` function."""

    def __init__(self):
        self.list = []

    def __deepcopy__(self, memo):
        other = OptimizeResult()
        other.list = deepcopy(self.list)
        return other

    def __getattr__(self, key):
        """Define `optimize_result.key`."""
        try:
            return [res[key] for res in self.list]
        except KeyError:
            raise AttributeError(key) from None

    def __getitem__(self, index):
        """Define `optimize_result[i]` to access the i-th result."""
        try:
            return self.list[index]
        except IndexError:
            raise IndexError(
                f"{index} out of range for optimize result of "
                f"length {len(self.list)}."
            ) from None

    def __getstate__(self):
        # while we override __getattr__ as we do now, this is required to keep
        # instances pickle-able
        return vars(self)

    def __setstate__(self, state):
        # while we override __getattr__ as we do now, this is required to keep
        # instances pickle-able
        vars(self).update(state)

    def __len__(self):
        return len(self.list)

    def summary(
        self,
        disp_best: bool = True,
        disp_worst: bool = False,
        full: bool = False,
        show_hess: bool = True,
    ) -> str:
        """
        Get summary of the object.

        Parameters
        ----------
        disp_best:
            Whether to display a detailed summary of the best run.
        disp_worst:
            Whether to display a detailed summary of the worst run.
        full:
            If True, print full vectors including fixed parameters.
        show_hess:
            If True, display the Hessian of the OptimizerResult.
        """
        if len(self) == 0:
            return "## Optimization Result \n\n*empty*\n"

        # perform clustering for better information
        clust, clustsize = assign_clusters(delete_nan_inf(self.fval)[1])

        # aggregate exit messages
        message_counts_df = pd.DataFrame(
            Counter(self.message).most_common(), columns=["Message", "Count"]
        )
        counter_message = message_counts_df[["Count", "Message"]].to_markdown(
            index=False
        )
        counter_message = "  " + counter_message.replace("\n", "\n  ")

        times_message = (
            f"\t* Mean execution time: {np.mean(self.time):0.3f}s\n"
            f"\t* Maximum execution time: {np.max(self.time):0.3f}s,"
            f"\tid={self[np.argmax(self.time)].id}\n"
            f"\t* Minimum execution time: {np.min(self.time):0.3f}s,\t"
            f"id={self[np.argmin(self.time)].id}"
        )

        # special handling in case there are only non-finite fvals
        num_best_value = int(clustsize[0]) if len(clustsize) else len(self)
        num_plateaus = (
            (1 + max(clust) - sum(clustsize == 1)) if len(clustsize) else 0
        )

        summary = (
            "## Optimization Result \n\n"
            f"* number of starts: {len(self)} \n"
            f"* best value: {self[0]['fval']}, id={self[0]['id']}\n"
            f"* worst value: {self[-1]['fval']}, id={self[-1]['id']}\n"
            f"* number of non-finite values: "
            f"{np.logical_not(np.isfinite(self.fval)).sum()}\n\n"
            f"* execution time summary:\n{times_message}\n"
            f"* summary of optimizer messages:\n\n{counter_message}\n\n"
            f"* best value found (approximately) {num_best_value} time(s)\n"
            f"* number of plateaus found: {num_plateaus}\n"
        )
        if disp_best:
            summary += (
                f"\nA summary of the best run:\n\n"
                f"{self[0].summary(full, show_hess=show_hess)}"
            )
        if disp_worst:
            summary += (
                f"\nA summary of the worst run:\n\n{self[-1].summary(full)}"
            )
        return summary

    def append(
        self,
        optimize_result: OptimizationResult | list[OptimizerResult],
        sort: bool = True,
        prefix: str = "",
    ):
        """
        Append an OptimizerResult or an OptimizeResult to the result object.

        Parameters
        ----------
        optimize_result:
            The result of one or more (local) optimizer run.
        sort:
            Boolean used so we only sort once when appending an
            optimize_result.
        prefix:
            The IDs for all appended results will be prefixed with this.
        """
        current_ids = set(self.id)
        if isinstance(optimize_result, OptimizeResult | list):
            result_list = (
                optimize_result.list
                if isinstance(optimize_result, OptimizeResult)
                else optimize_result
            )
            identifiers = [
                optimizer_result.id for optimizer_result in result_list
            ]
            new_ids = {
                prefix + identifier
                for identifier in identifiers
                if identifier is not None
            }
            if not current_ids.isdisjoint(new_ids):
                raise ValueError(
                    "Some IDs you want to merge coincide with "
                    f"the existing IDs: {current_ids & new_ids}. "
                    "Please use an appropriate prefix such as 'run_2_'."
                )
            for optimizer_result in result_list:
                if optimizer_result.id is not None:
                    optimizer_result.id = prefix + optimizer_result.id
            self.list.extend(result_list)
        elif isinstance(optimize_result, OptimizerResult):
            # if id is None, append without checking for duplicate ids
            if optimize_result.id is None:
                self.list.append(optimize_result)
            else:
                new_id = prefix + optimize_result.id
                if new_id in current_ids:
                    raise ValueError(
                        f"The id `{new_id}` you want to merge coincides with "
                        "the existing id's. Please use an "
                        "appropriate prefix such as 'run_2_'."
                    )
                optimize_result.id = new_id
                self.list.append(optimize_result)
        else:
            raise ValueError(
                "Argument `optimize_result` is of unsupported "
                f"type {type(optimize_result)}."
            )
        if sort:
            self.sort()

    def sort(self):
        """Sort the optimizer results by function value fval (ascending)."""

        def get_fval(res):
            return fval if not np.isnan(fval := res.fval) else np.inf

        self.list = sorted(self.list, key=get_fval)

    def as_dataframe(self, keys=None) -> pd.DataFrame:
        """
        Get as pandas DataFrame.

        If keys is a list, return only the specified values, otherwise all.
        """
        lst = self.as_list(keys)

        df = pd.DataFrame(lst)

        return df

    def as_list(self, keys=None) -> Sequence:
        """
        Get as list.

        If keys is a list, return only the specified values.

        Parameters
        ----------
        keys: list(str), optional
            Labels of the field to extract.
        """
        lst = self.list

        if keys is not None:
            lst = [{key: res[key] for key in keys} for res in lst]

        return lst

    def get_for_key(self, key) -> list:
        """Extract the list of values for the specified key as a list."""
        warnings.warn(
            "get_for_key() is deprecated in favour of "
            "optimize_result.key and will be removed in future "
            "releases.",
            DeprecationWarning,
            stacklevel=1,
        )
        return [res[key] for res in self.list]

    def get_by_id(self, ores_id: str):
        """Get OptimizationResult with the specified id."""
        for res in self.list:
            if res.id == ores_id:
                return res
        else:
            raise ValueError(f"no optimization result with id={ores_id}")


class LazyOptimizerResult(OptimizerResult):
    """
    A class to handle lazy loading of optimizer results from an HDF5 file.

    This class extends the OptimizerResult class and overrides methods to
    load data only when it is accessed, improving memory usage and performance
    for large datasets.

    Attributes
    ----------
    filename : str
        The path to the HDF5 file containing the optimizer results.
    group_name : str
        The name of the group in the HDF5 file where the results are stored.
    with_history : bool
        Whether to load the optimization history when accessed.
    _data : dict
        A dictionary to store loaded data.
    _metadata_loaded : bool
        A flag indicating whether metadata has been loaded.
    """

    def __init__(self, filename, group_name):
        """
        Initialize a LazyOptimizerResult instance.

        Parameters
        ----------
        filename : str
            The path to the HDF5 file containing the optimizer results.
        group_name : str
            The name of the group in the HDF5 file where the results are stored.
        with_history : bool
            Whether to load the optimization history when accessed.
        """
        # Initialize parent OptimizerResult (which inherits from dict)
        super().__init__()
        # Store these attributes in __dict__ instead of as dict items
        # This avoids conflicts with the lazy loading mechanism
        self.__dict__["filename"] = filename
        self.__dict__["group_name"] = group_name
        self.__dict__["_data"] = {}

    def _get_value(self, key):
        """
        Get the value of a key.

        Parameters
        ----------
        key : str
            The key to get the value of.

        Returns
        -------
        value
            The value of the key.
        """
        _data = self.__dict__["_data"]

        if key not in _data:
            filename = self.__dict__["filename"]
            group_name = self.__dict__["group_name"]

            with h5py.File(filename, "r") as f:
                if key in f[group_name]:
                    _data[key] = f[f"{group_name}/{key}"][()]
                elif key in f[group_name].attrs:
                    _data[key] = f[group_name].attrs[key]
                else:
                    raise AttributeError(f"{key} not found in the HDF5 file.")
        return _data[key]

    def __getitem__(self, key):
        """
        Enable dictionary-style access to lazy-loaded attributes.

        Parameters
        ----------
        key : str
            The key to access.

        Returns
        -------
        value
            The value associated with the key.
        """
        try:
            return self._get_value(key)
        except AttributeError as e:
            raise KeyError(str(e)) from None

    @property
    def id(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("id")

    @property
    def x(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("x")

    @x.setter
    def x(self, value):
        """Setter for the x property."""
        self._data["x"] = value

    @property
    def fval(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("fval")

    @fval.setter
    def fval(self, value):
        """Setter for the fval property."""
        self._data["fval"] = value

    @property
    def grad(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("grad")

    @grad.setter
    def grad(self, value):
        """Setter for the grad property."""
        self._data["grad"] = value

    @property
    def hess(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("hess")

    @hess.setter
    def hess(self, value):
        """Setter for the hess property."""
        self._data["hess"] = value

    @property
    def res(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("res")

    @res.setter
    def res(self, value):
        """Setter for the res property."""
        self._data["res"] = value

    @property
    def sres(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("sres")

    @sres.setter
    def sres(self, value):
        """Setter for the sres property."""
        self._data["sres"] = value

    @property
    def n_fval(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("n_fval")

    @n_fval.setter
    def n_fval(self, value):
        """Setter for the n_fval property."""
        self._data["n_fval"] = value

    @property
    def n_grad(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("n_grad")

    @n_grad.setter
    def n_grad(self, value):
        """Setter for the n_grad property."""
        self._data["n_grad"] = value

    @property
    def n_hess(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("n_hess")

    @n_hess.setter
    def n_hess(self, value):
        """Setter for the n_hess property."""
        self._data["n_hess"] = value

    @property
    def n_res(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("n_res")

    @n_res.setter
    def n_res(self, value):
        """Setter for the n_res property."""
        self._data["n_res"] = value

    @property
    def n_sres(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("n_sres")

    @n_sres.setter
    def n_sres(self, value):
        """Setter for the n_sres property."""
        self._data["n_sres"] = value

    @property
    def x0(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("x0")

    @x0.setter
    def x0(self, value):
        """Setter for the x0 property."""
        self._data["x0"] = value

    @property
    def fval0(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("fval0")

    @fval0.setter
    def fval0(self, value):
        """Setter for the fval0 property."""
        self._data["fval0"] = value

    @property
    def history(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("history")

    @history.setter
    def history(self, value):
        """Setter for the history property."""
        self._data["history"] = value

    @property
    def exitflag(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("exitflag")

    @exitflag.setter
    def exitflag(self, value):
        """Setter for the exitflag property."""
        self._data["exitflag"] = value

    @property
    def time(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("time")

    @time.setter
    def time(self, value):
        """Setter for the time property."""
        self._data["time"] = value

    @property
    def message(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("message")

    @message.setter
    def message(self, value):
        """Setter for the message property."""
        self._data["message"] = value

    @property
    def optimizer(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("optimizer")

    @optimizer.setter
    def optimizer(self, value):
        """Setter for the optimizer property."""
        self._data["optimizer"] = value

    @property
    def free_indices(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("free_indices")

    @free_indices.setter
    def free_indices(self, value):
        """Setter for the free_indices property."""
        self._data["free_indices"] = value

    @property
    def inner_parameters(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("inner_parameters")

    @inner_parameters.setter
    def inner_parameters(self, value):
        """Setter for the inner_parameters property."""
        self._data["inner_parameters"] = value

    @property
    def spline_knots(self):
        """See :class:`OptimizerResult`."""
        return self._get_value("spline_knots")

    @spline_knots.setter
    def spline_knots(self, value):
        """Setter for the spline_knots property."""
        self._data["spline_knots"] = value



from __future__ import annotations

import numpy as np
import pandas as pd

from ...C import (
    CURRENT_SIMULATION,
    DATAPOINTS,
    EXPDATA_MASK,
    INNER_NOISE_PARS,
    INNER_PARAMETER_BOUNDS,
    LIN,
    MAX_DATAPOINT,
    MEASUREMENT_TYPE,
    MIN_DATAPOINT,
    N_SPLINE_PARS,
    NUM_DATAPOINTS,
    OPTIMIZE_NOISE,
    PARAMETER_TYPE,
    SEMIQUANTITATIVE,
    SPLINE_PAR_TYPE,
    TIME,
    InnerParameterType,
)
from ..base_problem import (
    AmiciInnerProblem,
    _get_timepoints_with_replicates,
    ix_matrices_from_arrays,
    scale_value,
)
from .parameter import SplineInnerParameter

try:
    import amici.sim.sundials as asd
    import petab.v1 as petab
    from petab.v1.C import (
        ESTIMATE,
        LOWER_BOUND,
        NOISE_PARAMETERS,
        OBSERVABLE_ID,
        PARAMETER_ID,
        PARAMETER_SCALE,
        UPPER_BOUND,
    )
except ImportError:
    pass


class SemiquantProblem(AmiciInnerProblem):
    r"""Inner optimization problem for semi-quantitative data.

    The inner problem for semi-quantitative data consists of spline parameters
    and noise parameters for semi-quantitative observables. The unknown
    nonlinear measurement mapping is estimated using a piece-wise linear spline.

    Attributes
    ----------
    xs:
        Mapping of (inner) parameter ID to ``InnerParameters``.
    data:
        Measurement data. One matrix (`num_timepoints` x `num_observables`)
        per simulation condition. Missing observations as NaN.
    edatas:
        AMICI ``ExpData``\s for each simulation condition.
    groups:
        A dictionary of the groups of the subproblem.
    spline_ratio:
        The ratio of the number of spline inner parameters and number of measurements for each group.
    """

    def __init__(
        self,
        spline_ratio: float = 0.5,
        **kwargs,
    ):
        """Construct."""
        super().__init__(**kwargs)
        self.spline_ratio = spline_ratio

        if spline_ratio <= 0:
            raise ValueError("Spline ratio must be a positive float.")
        self._initialize_groups()

    def _initialize_groups(self) -> None:
        """Initialize the groups of the subproblem."""
        self.groups = {}
        for group in self.get_groups_for_xs(InnerParameterType.SPLINE):
            xs = self.get_xs_for_group(group)

            self.groups[group] = {}
            self.groups[group][N_SPLINE_PARS] = len({x.index for x in xs})
            self.groups[group][DATAPOINTS] = self.get_measurements_for_group(
                group
            )
            self.groups[group][NUM_DATAPOINTS] = len(
                self.groups[group][DATAPOINTS]
            )
            self.groups[group][MIN_DATAPOINT] = np.min(
                self.groups[group][DATAPOINTS]
            )
            self.groups[group][MAX_DATAPOINT] = np.max(
                self.groups[group][DATAPOINTS]
            )

            self.groups[group][EXPDATA_MASK] = xs[0].ixs
            self.groups[group][CURRENT_SIMULATION] = np.zeros(
                self.groups[group][NUM_DATAPOINTS]
            )
            self.groups[group][INNER_NOISE_PARS] = 1
            self.groups[group][OPTIMIZE_NOISE] = (
                len(self.get_noise_parameters_for_group(group)) > 0
            )

    def initialize(self) -> None:
        """Initialize the subproblem."""
        # Initialize all parameter values.
        for x in self.xs.values():
            x.initialize()

        # Initialize the groups.
        for group in self.get_groups_for_xs(InnerParameterType.SPLINE):
            self.groups[group][CURRENT_SIMULATION] = np.zeros(
                self.groups[group][NUM_DATAPOINTS]
            )
            self.groups[group][INNER_NOISE_PARS] = 1

    @staticmethod
    def from_petab_amici(
        petab_problem: petab.Problem,
        amici_model: asd.Model,
        edatas: list[asd.ExpData],
        spline_ratio: float = None,
    ) -> SemiquantProblem:
        """Construct the inner problem from the `petab_problem`."""
        if spline_ratio is None:
            spline_ratio = get_default_options()
        return spline_inner_problem_from_petab_problem(
            petab_problem, amici_model, edatas, spline_ratio
        )

    def get_interpretable_x_ids(self) -> list[str]:
        """Get IDs of interpretable inner parameters.

        The interpretable inner parameters of the semiquantitative
        problem are the noise parameters.
        """
        return [
            x.inner_parameter_id
            for x in self.xs.values()
            if x.inner_parameter_type == InnerParameterType.SIGMA
        ]

    def get_interpretable_x_scales(self) -> list[str]:
        """Get scales of interpretable inner parameters.

        The interpretable inner parameters of the semiquantitative
        problem are the noise parameters.
        """
        return [
            x.scale
            for x in self.xs.values()
            if x.inner_parameter_type == InnerParameterType.SIGMA
        ]

    def get_semiquant_observable_ids(self) -> list[str]:
        """Get the IDs of semiquantitative observables."""
        return list(
            {
                x.observable_id
                for x in self.xs.values()
                if x.inner_parameter_type == InnerParameterType.SPLINE
            }
        )

    def get_groups_for_xs(self, inner_parameter_type: str) -> list[int]:
        """Get unique list of ``SplineParameter.group`` values."""
        groups = [x.group for x in self.get_xs_for_type(inner_parameter_type)]
        return list(set(groups))

    def get_xs_for_group(self, group: int) -> list[SplineInnerParameter]:
        r"""Get ``SplineParameter``\s that belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group
            and x.inner_parameter_type == InnerParameterType.SPLINE
        ]

    def get_free_xs_for_group(self, group: int) -> list[SplineInnerParameter]:
        r"""Get ``SplineParameter``\s that are free and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group
            and x.estimate is True
            and x.inner_parameter_type == InnerParameterType.SPLINE
        ]

    def get_fixed_xs_for_group(self, group: int) -> list[SplineInnerParameter]:
        r"""Get ``SplineParameter``\s that are fixed and belong to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group
            and x.estimate is False
            and x.inner_parameter_type == InnerParameterType.SPLINE
        ]

    def get_inner_noise_parameters(self) -> list[float]:
        """Get a list with all noise parameter values."""
        return [
            x.value for x in self.get_xs_for_type(InnerParameterType.SIGMA)
        ]

    def get_noise_parameters_for_group(
        self, group: int
    ) -> SplineInnerParameter:
        r"""Get the ``SplineParameter``\ that is a noise parameters and belongs to the given group."""
        return [
            x
            for x in self.xs.values()
            if x.group == group
            and x.inner_parameter_type == InnerParameterType.SIGMA
        ]

    def get_inner_parameter_dictionary(self) -> dict:
        """Get a dictionary with all inner parameter ids and their values."""
        inner_par_dict = {}
        for x_id, x in self.xs.items():
            inner_par_dict[x_id] = x.value
        return inner_par_dict

    def get_spline_knots(
        self,
    ) -> list[list[np.ndarray[float], np.ndarray[float]]]:
        """Get spline knots of all semiquantitative observables.

        Returns
        -------
        list[list[np.ndarray[float], np.ndarray[float]]]
            A list of lists with two arrays. Each list in the first level corresponds
            to a semiquantitative observable. Each of these lists contains two arrays:
            the first array contains the spline bases, the second array contains the
            spline knot values. The ordering of the observable lists is the same
            as in `pypesto.problem.hierarchical.semiquant_observable_ids`.
        """
        # We need the solver only for the rescaling function.
        from .solver import SemiquantInnerSolver

        all_spline_knots = []

        for group in self.get_groups_for_xs(InnerParameterType.SPLINE):
            group_dict = self.groups[group]
            n_spline_pars = group_dict[N_SPLINE_PARS]
            n_data_points = group_dict[NUM_DATAPOINTS]

            inner_pars = np.array(
                [x.value for x in self.get_xs_for_group(group)]
            )

            # Utility matrix for the spline knot calculation
            lower_trian = np.tril(np.ones((n_spline_pars, n_spline_pars)))
            knot_values = np.dot(lower_trian, inner_pars)

            _, knot_bases, _ = SemiquantInnerSolver._rescale_spline_bases(
                sim_all=group_dict[CURRENT_SIMULATION],
                N=n_spline_pars,
                K=n_data_points,
            )

            spline_knots_for_observable = [knot_bases, knot_values]

            all_spline_knots.append(spline_knots_for_observable)

        return all_spline_knots

    def get_measurements_for_group(self, gr) -> np.ndarray:
        """Get measurements for a group."""
        # Taking the ixs of first inner parameter since
        # all of them are the same for the same group.
        ixs = self.get_xs_for_group(gr)[0].ixs

        return np.concatenate(
            [
                self.data[condition_index][ixs[condition_index]]
                for condition_index in range(len(ixs))
            ]
        )

    def get_noise_dummy_values(self, scaled: bool) -> dict[str, float]:
        """Get dummy values for noise parameters of the semiquantitative observable."""
        return {
            x_id: scale_value(x.value, x.scale) if scaled else x.value
            for x_id, x in self.xs.items()
            if x.inner_parameter_type == InnerParameterType.SIGMA
        }


def get_default_options() -> dict:
    """Return the default spline problem options dictionary."""
    spline_ratio = 1 / 2
    return spline_ratio


def spline_inner_problem_from_petab_problem(
    petab_problem: petab.Problem,
    amici_model: asd.Model,
    edatas: list[asd.ExpData],
    spline_ratio: float = None,
):
    """Construct the inner problem from the `petab_problem`."""
    if spline_ratio is None:
        spline_ratio = get_default_options()
    elif spline_ratio <= 0:
        raise ValueError("Spline ratio must be a positive float.")

    # inner parameters
    inner_parameters = spline_inner_parameters_from_measurement_df(
        petab_problem.measurement_df, spline_ratio, amici_model
    )

    # noise parameters for semiquantitative observables
    noise_parameters = noise_inner_parameters_from_parameter_df(
        petab_problem, amici_model
    )
    inner_parameters.extend(noise_parameters)

    # used indices for all measurement specific parameters
    ixs = spline_ixs_for_measurement_specific_parameters(
        petab_problem, amici_model, inner_parameters
    )

    # transform experimental data
    data = [asd.ExpDataView(edata)["measurements"] for edata in edatas]

    # matrixify
    ix_matrices = ix_matrices_from_arrays(ixs, data)

    # assign matrices
    for par in inner_parameters:
        par.ixs = ix_matrices[par.inner_parameter_id]

    return SemiquantProblem(
        xs=inner_parameters,
        data=data,
        edatas=edatas,
        spline_ratio=spline_ratio,
    )


def spline_inner_parameters_from_measurement_df(
    df: pd.DataFrame,
    spline_ratio: float,
    amici_model: asd.Model,
) -> list[SplineInnerParameter]:
    """Create list of inner free spline parameters from PEtab measurement table."""
    df = df.reset_index()

    observable_ids = amici_model.get_observable_ids()

    par_type = SPLINE_PAR_TYPE
    estimate = True
    lb, ub = INNER_PARAMETER_BOUNDS[InnerParameterType.SPLINE].values()

    inner_parameters = []

    # Select the semiquantitative measurements.
    df = df[df[MEASUREMENT_TYPE] == SEMIQUANTITATIVE]

    # Iterate over groups.
    for observable_id in observable_ids:
        group = observable_ids.index(observable_id) + 1
        df_for_group = df[df[OBSERVABLE_ID] == observable_id]

        n_spline_parameters = int(np.ceil(len(df_for_group) * spline_ratio))

        # Create n_spline_parameters number of spline inner parameters.
        for par_index in range(n_spline_parameters):
            par_id = f"{par_type}_{observable_id}_{group}_{par_index + 1}"
            inner_parameters.append(
                SplineInnerParameter(
                    inner_parameter_id=par_id,
                    inner_parameter_type=InnerParameterType.SPLINE,
                    scale=LIN,
                    lb=lb,
                    ub=ub,
                    observable_id=observable_id,
                    group=group,
                    index=par_index + 1,
                    estimate=estimate,
                )
            )

    inner_parameters.sort(key=lambda x: (x.group, x.index))

    return inner_parameters


def noise_inner_parameters_from_parameter_df(
    petab_problem: petab.Problem,
    amici_model: asd.Model,
) -> list[SplineInnerParameter]:
    """Create list of inner free noise parameters from PEtab parameter table."""
    # Select the semiquantitative measurements.
    measurement_df = petab_problem.measurement_df
    measurement_df = measurement_df[
        measurement_df[MEASUREMENT_TYPE] == SEMIQUANTITATIVE
    ]

    observable_ids = amici_model.get_observable_ids()

    # Create a dictionary with unique pairs of observable id
    # and noise parameter from the measurement table.
    noise_parameter_to_observable = {}
    for _, row in measurement_df.iterrows():
        observable_id = row[OBSERVABLE_ID]
        noise_parameter_id = row[NOISE_PARAMETERS]
        noise_parameter_to_observable[noise_parameter_id] = observable_id

    noise_parameters = []

    parameter_df = petab_problem.parameter_df.reset_index()
    # Create noise inner parameters.
    for _, row in parameter_df.iterrows():
        if row[PARAMETER_ID] not in noise_parameter_to_observable:
            continue
        if petab.is_empty(row.get(PARAMETER_TYPE)) or not row[ESTIMATE]:
            continue
        observable_id = noise_parameter_to_observable[row[PARAMETER_ID]]
        group = observable_ids.index(observable_id) + 1

        noise_parameters.append(
            SplineInnerParameter(
                inner_parameter_id=row[PARAMETER_ID],
                inner_parameter_type=InnerParameterType.SIGMA,
                scale=row[PARAMETER_SCALE],
                lb=row[LOWER_BOUND],
                ub=row[UPPER_BOUND],
                observable_id=observable_id,
                group=group,
                index=None,
                estimate=True,
            )
        )

    return noise_parameters


def spline_ixs_for_measurement_specific_parameters(
    petab_problem: petab.Problem,
    amici_model: asd.Model,
    inner_parameters: list[SplineInnerParameter],
) -> dict[str, list[tuple[int, int, int]]]:
    """Create mapping of parameters to measurements.

    Returns
    -------
    A dictionary mapping parameter ID to a list of
    `(condition index, time index, observable index)` tuples in which this
    output parameter is used. For each condition, the time index refers to
    a sorted list of non-unique time points for which there are measurements.
    """
    ixs_for_par = {}
    observable_ids = amici_model.get_observable_ids()

    simulation_conditions = (
        petab_problem.get_simulation_conditions_from_measurement_df()
    )
    for condition_ix, condition in simulation_conditions.iterrows():
        # measurement table for current condition
        df_for_condition = petab.get_rows_for_condition(
            measurement_df=petab_problem.measurement_df, condition=condition
        )

        # unique sorted list of timepoints
        timepoints = sorted(df_for_condition[TIME].unique().astype(float))
        # non-unique sorted list of timepoints
        timepoints_w_reps = _get_timepoints_with_replicates(
            measurement_df=df_for_condition
        )

        for time in timepoints:
            # subselect measurements for time `time`
            df_for_time = df_for_condition[df_for_condition[TIME] == time]
            time_ix_0 = timepoints_w_reps.index(time)

            # remember used time indices for each observable
            time_ix_for_obs_ix = {}

            # iterate over measurements
            for _, measurement in df_for_time.iterrows():
                # extract observable index
                observable_ix = observable_ids.index(
                    measurement[OBSERVABLE_ID]
                )

                # as the time indices have to account for replicates, we need
                #  to track which time indices have already been assigned for
                #  the current observable
                if observable_ix in time_ix_for_obs_ix:
                    # a replicate
                    time_ix_for_obs_ix[observable_ix] += 1
                else:
                    # the first measurement for this `(observable, timepoint)`
                    time_ix_for_obs_ix[observable_ix] = time_ix_0
                time_w_reps_ix = time_ix_for_obs_ix[observable_ix]

                inner_par_ids_for_meas = (
                    get_spline_inner_par_ids_for_measurement(
                        measurement, inner_parameters
                    )
                )

                # try to insert if hierarchical parameter
                for override in inner_par_ids_for_meas:
                    ixs_for_par.setdefault(override, []).append(
                        (condition_ix, time_w_reps_ix, observable_ix)
                    )
    return ixs_for_par


def get_spline_inner_par_ids_for_measurement(
    measurement: dict,
    inner_parameters: list[SplineInnerParameter],
):
    """Return inner parameter ids of parameters which are related to the measurement."""
    return [
        inner_par.inner_parameter_id
        for inner_par in inner_parameters
        if inner_par.observable_id == measurement[OBSERVABLE_ID]
    ]
